/*
  $Id: write.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "atgpib.h"


static unsigned char nff;


static int board_dma_write( Board_t * board );
static int board_pio_write( Board_t * board );


/*---------------------------------------------------------------------------*
 * If interrupts can be used a down() is done on the transfer semaphore -
 * if the interrupt has already come in this does nothing because then up()
 * has already been called on the semaphore by the interrupt handler, other-
 * wise the process will sleep until the interrupt handler is triggered. If
 * no interrupts can be used the ISR3 and ISR1 registers are polled until
 * a new byte can be written or a timeout or error occured.
 *---------------------------------------------------------------------------*/

inline static void board_wait_out( Board_t * board )
{
	unsigned char isr3;


	PDEBUGin( );

	if ( board->status & USE_IRQ ) {

		/* Enable interrupt on "not full FIFO" condition, i.e. if
		   further data can be written to the FIFO */

		regs.imr3 |= IMR3_NFF_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		/* Wait until semaphore gets up()'ed by the interrupt or
		   timer handler */

		if ( ! board->timeout )
			down( &board->transf.sem );

		/* Disable interrupt on "not full FIFO" condition */

		regs.imr3 &= ~ IMR3_NFF_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		/* Get state and error bits */

		isr3 = board_in( reg_addr.ISR3 );
		nff  = isr3 & ISR3_NFF;
		done = isr3 & ISR3_DONE;
		err  = board_in( reg_addr.ISR1 ) & ISR1_ERR;
	} else
		/* Poll the relevant registers until byte can be written
		   or an error or timeout happened */

		do {
			isr3 = board_in( reg_addr.ISR3 );
			nff  = isr3 & ISR3_NFF;
			done = isr3 & ISR3_DONE;
			err  = board_in( reg_addr.ISR1 ) & ISR1_ERR;
		} while ( ! ( nff || done || err || board->timeout ) );

	PDEBUGout( );
}


/*------------------------------------------------------------------*
 * Function decides if the transfer is to be done using DMA or PIO.
 *------------------------------------------------------------------*/

int board_write( Board_t * board )
{
	if ( board->status & USE_DMA && board->transf.cnt > 1 )
		return board_dma_write( board );
	else
		return board_pio_write( board );
}


/*-------------------------------------------------------------------*
 * Function writes out the bytes stored in the transfer buffer in
 * the board structure using DMA. Also the number of bytes is stored
 * in the board structure as well as the information if this is the
 * last of a series of write operations to the same device. The
 * function expects the device to be addressed as listener and the
 * board as talker and the ATN line to be unasserted.
 *-------------------------------------------------------------------*/

static int board_dma_write( Board_t * board )
{
	unsigned int count;
	int ccount;


	PDEBUGin( );

	/* Clear ISR0, ISR1 and ISR2 register and FIFO */

	board_in( reg_addr.ISR0 );
	board_in( reg_addr.ISR1 );
	board_in( reg_addr.ISR2 );
	board_out( reg_addr.CMDR, RESET_FIFO );

	/* Setup board to write data to the bus in 16 bit transfer mode. Also
	   make sure the EOI line is raised with the last byte to be sent when
	   this is the last chunk of data and sending EOI is required. */

	regs.cfg |= CFG_TLCHLTE | CFG_16_8N;
	regs.cfg &= ~ ( CFG_CMD | CFG_IN );
	if ( board->transf.is_last && board->eosmode & GPIB_EOT )
		regs.cfg |= CFG_CCEN;
	else
		regs.cfg &= ~ CFG_CCEN;
	board_out( reg_addr.CFG, regs.cfg );

	/* Write 2's complement of length of the data buffer to be written
	   into the count registers */

	ccount = - board->transf.cnt;
	board_out( reg_addr.CNT0,
		   ( unsigned char ) ( ccount & 0xFF ) );
	board_out( reg_addr.CNT1,
		   ( unsigned char ) ( ( ccount >>  8 ) & 0xFF ) );
	board_out( reg_addr.CNT2,
		   ( unsigned char ) ( ( ccount >> 16 ) & 0xFF ) );
	board_out( reg_addr.CNT3,
		   ( unsigned char ) ( ( ccount >> 24 ) & 0xFF ) );

	/* Enable board side DMA */

	if ( ! ( regs.accwr & ACCWR_DMAEN ) ) {
		regs.accwr |= ACCWR_DMAEN;
		board_out( reg_addr.ACCWR, regs.accwr );
	}

	/* Clear IMR0 and IMR2 */

	if ( regs.imr0 & 0x7F ) {
		regs.imr0 = 0x80;
		board_out( reg_addr.IMR0, regs.imr0 );
	}

	if ( regs.imr2 ) {
		regs.imr2 = 0;
		board_out( reg_addr.IMR2, regs.imr2 );
	}

	/* Enable DMA write via the FIFO using 16-bit mode (take care: the
	   second argument is the number of bytes to be written which must
	   be an even number, so better tell it to write one byte too many
	   if there's an odd number of bytes) */

	gpib_isa_dma_enable( board->transf.buf,
			     ( board->transf.cnt + 1 ) & 0xfffe, WRITE );

	/* Set up the transfer semaphore and enable interrupts to make the
	   board flag errors and DONE state */

	semaphore_init( &board->transf.sem, 0 );

	regs.imr1 |= IMR1_ERR_IE;
	board_out( reg_addr.IMR1, regs.imr1 );

	regs.imr3 |= IMR3_TLCINT_IE | IMR3_DONE_IE;
	board_out( reg_addr.IMR3, regs.imr3 );

	regs.intr |= INTR_INTEN;
	board_out( reg_addr.INTR, regs.intr );

	/* Send the GO command to initiate board side transfer */

	board_out( reg_addr.CMDR, GO );

	/* Wait for transfer to finish or an interrupt or timeout to occur */

	if ( ! board->timeout )
		down( &board->transf.sem );

	/* Send STOP command and reset command bit in CFG register */

	board_out( reg_addr.CMDR, STOP );

	/* Disable host side DMA */

	gpib_isa_dma_disable( );

	/* Disable DMA and interrupts */

	if ( regs.accwr & ACCWR_DMAEN ) {
		regs.accwr &= ~ ACCWR_DMAEN;
		board_out( reg_addr.ACCWR, regs.accwr );
	}

	regs.intr &= ~ INTR_INTEN;
	board_out( reg_addr.INTR, regs.intr );

	regs.imr3 &= ~ ( IMR3_TLCINT_IE | IMR3_DONE_IE );
	board_out( reg_addr.IMR3, regs.imr3 );

	/* Reset CFG and ISR1 registers to normal state */

	regs.cfg &= ~ CFG_TLCHLTE;
	board_out( reg_addr.CFG, regs.cfg );

	regs.imr1 &= ~ IMR1_ERR_IE;
	board_out( reg_addr.IMR1, regs.imr1 );

	/* Calculate how many bytes have been written */

	count = - (   ( board_in( reg_addr.CNT0 ) & 0xFF )	   |
		    ( ( board_in( reg_addr.CNT1 ) & 0xFF ) <<  8 ) |
		    ( ( board_in( reg_addr.CNT2 ) & 0xFF ) << 16 ) |
		    ( ( board_in( reg_addr.CNT3 ) & 0xFF ) << 24 ) );
	gpib_count = board->transf.cnt - count;

	PDEBUG( "Wrote %ld of %d bytes\n", gpib_count, board->transf.cnt );

	/* Set the result flags */

	if ( board_in( reg_addr.ISR1 ) & ISR1_ERR ) {
		gpib_status |= GPIB_ERR;
		if ( board->timeout )
			gpib_status |= GPIB_TIMO;
		gpib_error   = GPIB_ENOL;
		PDEBUG( "No listener in DMA write\n" );
	} else if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
		PDEBUG( "Timeout in DMA write\n" );
	} else if ( board->transf.is_last && count == 0 )
		gpib_status |= GPIB_CMPL;

	PDEBUGout( );
	return gpib_status;
}


/*--------------------------------------------------------------------*
 * Function writes out the bytes stored in the transfer buffer in the
 * board structure in PIO mode. Also the number of bytes is stored in
 * the board structure as well as the information if this is the last
 * of a series of write operations to the same device. The function
 * expects the device to be addressed as listener and the board as
 * talker and the ATN line to be unasserted.
 *--------------------------------------------------------------------*/

static int board_pio_write( Board_t * board )
{
	unsigned int count = 0;
	int ccount;


	PDEBUGin( );

	/* Clear ISR0, ISR1 and ISR2 register and FIFO */

	board_in( reg_addr.ISR0 );
	board_in( reg_addr.ISR1 );
	board_in( reg_addr.ISR2 );
	board_out( reg_addr.CMDR, RESET_FIFO );

	/* Setup board to write data to the bus in 8 bit transfer mode and make
	   sure the EOI line is raised with the last byte to be sent when we're
	   at the last chunk of data and sending EOI is supposed to happen.
	   Writes happen only via FIFO B. */

	regs.cfg |= CFG_TLCHLTE;
	regs.cfg &= ~ ( CFG_CMD | CFG_IN | CFG_16_8N | CFG_A_BN );
	if ( board->transf.is_last && board->eosmode & GPIB_EOT )
		regs.cfg |= CFG_CCEN;
	else
		regs.cfg &= ~ CFG_CCEN;
	board_out( reg_addr.CFG, regs.cfg );

	/* Write 2's complement of length of the data buffer to be written
	   into the count registers */

	ccount = - board->transf.cnt;
	board_out( reg_addr.CNT0,
		   ( unsigned char ) ( ccount & 0xFF ) );
	board_out( reg_addr.CNT1,
		   ( unsigned char ) ( ( ccount >>  8 ) & 0xFF ) );
	board_out( reg_addr.CNT2,
		   ( unsigned char ) ( ( ccount >> 16 ) & 0xFF ) );
	board_out( reg_addr.CNT3,
		   ( unsigned char ) ( ( ccount >> 24 ) & 0xFF ) );

	/* Make sure no DMA is used and clear IMR0 and IMR2 */

	if ( regs.accwr & ACCWR_DMAEN ) {
		regs.accwr &= ~ ACCWR_DMAEN;
		board_out( reg_addr.ACCWR, regs.accwr );
	}

	if ( regs.imr0 & 0x7F ) {
		regs.imr0 = 0x80;
		board_out( reg_addr.IMR0, regs.imr0 );
	}

	if ( regs.imr2 ) {
		regs.imr2 = 0;
		board_out( reg_addr.IMR2, regs.imr2 );
	}

	/* Enable interrupts (if they're to be used) and set up the transfer
	   semaphore */

	if ( board->status & USE_IRQ ) {
		semaphore_init( &board->transf.sem, 0 );

		regs.imr1 |= IMR1_ERR_IE;
		board_out( reg_addr.IMR1, regs.imr1 );

		regs.imr3 |= IMR3_TLCINT_IE | IMR3_DONE_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		regs.intr |= INTR_INTEN;
		board_out( reg_addr.INTR, regs.intr );
	}

	/* Send the GO command */

	board_out( reg_addr.CMDR, GO );

	/* Write out all bytes to the FIFO using 8-bit PIO mode */

	done = err = 0;

	while ( count < board->transf.cnt ) {
		if ( ! ( nff = board_in( reg_addr.ISR3 ) & ISR3_NFF ) )
			board_wait_out( board );

		if ( done || err || board->timeout )
			goto wrt_done;

		board_out( reg_addr.FIFOB,
			   * ( board->transf.buf + count++ ) );
	}

	/* Wait for end of transfer when everything was written */

	while ( ! ( done || err || board->timeout ) )
		board_wait_done( board );

 wrt_done:

	/* Send STOP command and reset command bit in CFG register */

	board_out( reg_addr.CMDR, STOP );

	/* Disable the interrupts (if used) */

	if ( board->status & USE_IRQ ) {
		regs.intr &= ~ INTR_INTEN;
		board_out( reg_addr.INTR, regs.intr );

		regs.imr1 &= ~ IMR1_ERR_IE;
		board_out( reg_addr.IMR1, regs.imr1 );

		regs.imr3 &= ~ ( IMR3_NFF_IE | IMR3_TLCINT_IE | IMR3_DONE_IE );
		board_out( reg_addr.IMR3, regs.imr3 );
	}

	/* Reset CFG to normal state */

	regs.cfg &= ~ CFG_TLCHLTE;
	board_out( reg_addr.CFG, regs.cfg );

	/* Calculate how many bytes have been written to the bus */

	count = - (   ( board_in( reg_addr.CNT0 ) & 0xFF )	      |
		    ( ( board_in( reg_addr.CNT1 ) & 0xFF ) <<  8 ) |
		    ( ( board_in( reg_addr.CNT2 ) & 0xFF ) << 16 ) |
		    ( ( board_in( reg_addr.CNT3 ) & 0xFF ) << 24 ) );

	gpib_count = board->transf.cnt - count;

	PDEBUG( "board_pio_write: Wrote %ld of %d bytes\n",
		gpib_count, board->transf.cnt );

	/* Set the result flags */

	if ( err ) {
		gpib_status |= GPIB_ERR;
		if ( board->timeout )
			gpib_status |= GPIB_TIMO;
		gpib_error   = GPIB_ENOL;
		PDEBUG( "No listener in PIO write\n" );
	} else if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
		PDEBUG( "Timeout in PIO write\n" );
	} else if ( board->transf.is_last && count == 0 )
		gpib_status |= GPIB_CMPL;

	PDEBUGout( );
	return gpib_status;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
