/*
  $Id: read.c 687 2006-01-26 12:18:44Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "pcII.h"


static int board_dma_read( Board_t * board );
static int board_pio_read( Board_t * board );
static unsigned char board_wait_in( Board_t * board );



/*-------------------------------------------------------------------------*
 * Function decides if the transfer is to be done using DMA or PIO.
 * Unfortunately, it is not possible to resume a DMA transfer. Even worse,
 * it seems that the DI or DO bit in ISR1 doesn't get raised after a DMA
 * transfer when the device wasn't re-addressed. Thus transfers that have
 * to be split in several chunks due to the transfer buffer not being
 * large enough can only be done using PIO.
 *-------------------------------------------------------------------------*/

int board_read( Board_t * board )
{
	int retval;
	static bool cont = UNSET;             /* flag, set on split transfer */


	if ( ! board->transf.is_last )        /* is this a split transfer ? */
		cont = SET;

	/* DMA can only be used when it's allowed and for non-split
	   transfers */

	if ( board->status & USE_DMA && ! cont )
		retval = board_dma_read( board );
	else
		retval = board_pio_read( board );

	/* If this is the very end of a transfer reset 'const' */

	if ( board->transf.is_last || gpib_status & GPIB_ERR ||
	     gpib_status & GPIB_END )
		cont = UNSET;

	return retval;
}


/*----------------------------------------------------------------*
 * The function reads in using DMA as many bytes as fit into the
 * transfer buffer in the board structure or until an END signal
 * was received. The size of the transfer buffer as well as the
 * information if this is the last chunk of data to be read in a
 * series of read opreations. The function expects the device to
 * addressed as talker and the board as listener and the ATN line
 * to be unasserted.
 *----------------------------------------------------------------*/

static int board_dma_read( Board_t * board )
{
	int residue;
	bool failure = UNSET;
	unsigned char isr1;


	/* If there's only one byte PIO mode for output will do */

	if ( board->transf.cnt == 1 )
		return board_pio_read( board );

	PDEBUGin( );

	/* Enable DMA transfer */

	gpib_isa_dma_enable( board->transf.buf, board->transf.cnt, READ );

	/* 'Arm' the transfer semaphore and enable interrupt and DMA transfer
	   - I don't know what writing 0xff to the port ( 0x2f0 | board->irq )
	   actually does but it seems to be the only way to get the interrupt
	   working with the NI board with a NAT4882 chip in PCIIA mode (it
	   doesn't seem to help with a PCIIA board by GTI with a real
	   uPD7210) */

	board_in( reg_addr.ISR1 );
	semaphore_init( &board->transf.sem, 0 );
	if ( board->type == PCIIA )
		board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );
	board_out( reg_addr.IMR2, HR_DMAI );
	board_out( reg_addr.IMR1, HR_ERRIE | HR_ENDIE );

	/* Set mode to holdoff on END and remove current holdoff */

	board_set_eosmode( board, HR_HLDE );
	if ( board->status & HOLDOFF )
		board_out( reg_addr.AUXMR, AUX_rhdf );

	/* Wait for transfer to finish */

	down( &board->transf.sem );

	/* Disable further interrupts */

	board_out( reg_addr.IMR1, 0 );
	board_out( reg_addr.IMR2, 0 );
	if ( board->type == PCIIA )
		board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );

	/* Disable DMA transfer and get number of bytes read in */

	residue = gpib_isa_dma_disable( );
	gpib_count = board->transf.cnt - residue;

	/* Check for errors: A failure is either a timeout or an read error or
	   not all bytes could be read eventhough no END message was
	   received */

	isr1 = board_in( reg_addr.ISR1 );
	failure = board->timeout ||
	          ( residue != 0 && ! ( isr1 & HR_END ) );

	PDEBUG( "board->transf.cnt = %u, residue %u %s END\n",
		board->transf.cnt, residue, isr1 &
		HR_END ? "with" : "without" );

	/* Set the status flags */

	if ( failure ) {
		gpib_status |= GPIB_ERR;
		if ( board->timeout )
			gpib_status |= GPIB_TIMO;
		gpib_error = GPIB_EABO;
		if ( ! ( isr1 & HR_END ) )
			board->status &= ~HOLDOFF;
	} else  {
		if ( isr1 & HR_END || gpib_count == board->transf.cnt ) {
			gpib_status |= GPIB_END;
			board->status |= HOLDOFF;
		}
	}

	PDEBUGout( );
	return gpib_status;
}


/*----------------------------------------------------------------*
 * The function reads in using PIO as many bytes as fit into the
 * transfer buffer in the board structure or until an END signal
 * was received. The size of the transfer buffer as well as the
 * information if this is the last chunk of data to be read in a
 * series of read opreations. The function expects the device to
 * addressed as talker and the board as listener and the ATN line
 * to be unasserted.
 *----------------------------------------------------------------*/

static int board_pio_read( Board_t * board )
{
	unsigned int count = 0;
	unsigned char isr1;


	PDEBUGin( );

	/* If interrupts can be used the transfer semaphore must be 'armed'
	   and the interrupt must be enabled - I don't know what writing 0xff
	   to the port ( 0x2f0 | board->irq ) actually does but it seems to be
	   the only way to get the interrupt working with the NI board with
	   the NAT4882 chip in PCIIA mode (it doesn't seem to help with a
	   PCIIA board by GTI with a real uPD7210). */

	if ( board->status & USE_IRQ ) {
		semaphore_init( &board->transf.sem, 0 );
		if ( board->type == PCIIA )
			board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );
		board_out( reg_addr.IMR1, HR_ENDIE | HR_DIIE );
	}

	/* Set mode to holdoff on END and remove current holdoff */

	board_set_eosmode( board, HR_HLDE );
	if ( board->status & HOLDOFF )
		board_out( reg_addr.AUXMR, AUX_rhdf );

	/* Now read all bytes until the maximum count is reached, the END bit
	   is set or a timeout occurs */

	do {
		isr1 = board_wait_in( board );
		if ( board->timeout )
			break;
		( ( unsigned char * ) board->transf.buf )[ count++ ] =
		                                   board_in( reg_addr.DIR );
	} while ( count < board->transf.cnt && ! ( isr1 & HR_END ) );

	PDEBUG( "board->transf.cnt = %u, count %u %s END\n",
		board->transf.cnt, count,
		isr1 & HR_END ? "with" : "without" );

	/* Disable the interrupt */

	if ( board->status & USE_IRQ )
		board_out( reg_addr.IMR1, 0 );

	/* Set the result flags */

	if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
		if ( ! ( isr1 & HR_END ) )
			board->status &= ~HOLDOFF;
	} else {
		gpib_count = count;
		if ( isr1 & HR_END || gpib_count == board->transf.cnt ) {
			gpib_status |= GPIB_END;
			board->status |= HOLDOFF;
		}			
	}

	PDEBUGout( );
	return gpib_status;
}


/*---------------------------------------------------------------------------*
 * If interrupts can be used simply a down is done on the transfer semaphore
 * - if the interrupt has already come in this does nothing because then the
 * semaphore has been up'ed by the interrupt handler, otherwise the process
 * will sleep until the interrupt handler is triggered. If no interrupts are
 * used the ISR1 register is polled until a new data byte arrived in DIR or,
 * at least, the END condition is true (take care: on rare occasions the END
 * is set without DI being set, it seems as if the ED bit is set before the
 * DI and sometimes ISR1 gets read in in between...) or a timeout occurred.
 *---------------------------------------------------------------------------*/

static unsigned char board_wait_in( Board_t * board )
{
	unsigned char isr1 = 0;


	PDEBUGin( );

	if ( board->status & USE_IRQ ) {

		/* Don't do down() if timeout has occurred - the timeout might
		   have actually happend just before the semaphore was
		   initialized! */

		if ( ! board->timeout ) {
			down( &board->transf.sem );
			isr1 = board_in( reg_addr.ISR1 );
			if ( board->type == PCIIA )
				board_out( ( void * ) ( 0x2f0 | board->irq ),
					   0xff );
		}
	} else
		while ( ! ( ( isr1 = board_in( reg_addr.ISR1 ) ) &
			    ( HR_DI | HR_END ) ) &&
			! board->timeout )
			/* empty */ ;

	PDEBUGout( );
	return isr1;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
