/*
  $Id: cmd.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"



/*--------------------------------------------------------------*
 * Function for writing data onto the bus in command mode, i.e.
 * while the ATN line is raised - don't use it for transferring
 * data to a device, that's what gpib_write() is for!
 *--------------------------------------------------------------*/

int gpib_cmd( void __user * buffer,
	      unsigned long write_count )
{
	void __user *buf;
	unsigned long to_be_written,
		      already_written;
	unsigned int chunk_size;
	int ret_status;


	PDEBUGin( );

	/* Return if there's nothing actually to be written */

	if ( write_count == 0 ) {
		gpib_status |= GPIB_CMPL;
		PDEBUGout( );
		return 0;
	}

	/* Check if we have read access to the user supplied buffer */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	if ( access_ok( VERIFY_READ, buffer, write_count ) == 0 ) {
#else
	if ( verify_area( VERIFY_READ, buffer, write_count ) != 0 ) {
#endif
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUGout( );
		return - EFAULT;
	}

	/* Board must be active controller to be able to send commands */

	if ( ! ( gpib_status & GPIB_CIC && gpib_status & GPIB_ATN ) &&
	     ( ret_status = gpib_command_setup( ) ) != 0 ) {
		PDEBUGout( );
		return ret_status;
	}

	/* If the buffer isn't large enough to hold all the data to be written
	   we have to split the transfer into chunks of the buffers size */

	to_be_written = write_count;
	already_written = 0;
	buf = buffer;

	if ( to_be_written > board.buf_size ) {
		chunk_size = board.buf_size;
		board.transf.is_last = UNSET;
	} else {
		chunk_size = to_be_written;
		board.transf.is_last = SET;
	}

	/* Board must be active controller to be able to send commands */

	if ( ( ret_status = gpib_command_setup( ) ) != 0 ) {
		PDEBUGout( );
		return ret_status;
	}

	board.transf.buf = board.buf;

	/* Start the timer to avoid hanging if the device doesn't react */

	gpib_start_timer( );

	/* Now copy data from the user buffer and write them onto the bus,
	   repeating this until either all ata are written or an error did
	   occur */

	do {

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
		__copy_from_user( board.transf.buf, buf, chunk_size );
#else
		memcpy_fromfs( board.transf.buf, buf, chunk_size );
#endif

		board.transf.cnt = chunk_size;

		board_cmd( &board );

		/* Calculate how much still needs to be written */

		already_written += gpib_count;
		to_be_written -= gpib_count;

		/* Increment the pointer to user space to where the remaining
		   data are located */

		buf += gpib_count;
		gpib_count = 0;

		if ( to_be_written <= chunk_size ) {
			chunk_size = to_be_written;
			board.transf.is_last = SET;
		}

	} while ( to_be_written > 0 && ! ( gpib_status & GPIB_ERR ) );

	/* Stop the timer */

	gpib_stop_timer( );

	gpib_count = already_written;

	PDEBUGout( );
	return ( gpib_status & GPIB_ERR ) ? - EIO : 0;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
