/*
  $Id: init.c 735 2006-06-11 13:50:16Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"


char kernel_version[ ] = UTS_RELEASE;


/*-----------------------------------------------------------------------*
 * The following settings can be overwritten by options specified while
 * loading the module, otherwise the compiled in values are used
 * To overwrite the compiled-in settings call insmod (or set the options
 * in /etc/modules.conf) using a command line like this:
 *
 *   insmod gpib.o name=\"gpib0\" base=0x2b8 irq=5 dma=1
 *
 * Here is the complete list of module parameters:
 * name     - Identifier string to be used in output
 * base     - Memory base address of the board
 * irq      - IRQ level to be used (use -1 to disable use of interrupts)
 * dma      - DMA channel to be used (use -1 to disable DMA transfer)
 * major    - Major board device number (either 1-254 for a fixed major
 *            number or 0 to use a dynamically asigned major number)
 * minor    - Minor board device number, this number is simultaneously
 *            the primary address of the board (valid range is 0 to 30)
 * sad      - secondary GPIB address of board (or 0 to switch off
 *            secondary addressing of board, valid range is 96 to 126)
 * mem_size - Upper limit for amount of memory to be allocated for
 *            transfers, maximum is 64 kB
 * timo     - Indicates amount of time to wait before I/O or wait times
 *            out - valid values are 0 (no timeout) to 17 (100s)
 *-----------------------------------------------------------------------*/


char *name   = GPIB_DEF_NAME;       /* identifier and device file name */
int base     = GPIB_DEF_BASE;       /* IO memory address */
int irq      = GPIB_DEF_IRQ;        /* IRQ level */
int dma      = GPIB_DEF_DMA;        /* DMA channel */
int major    = GPIB_DEF_MAJOR;      /* major device number */
int minor    = GPIB_DEF_MINOR;      /* primary GPIB address */
int sad      = GPIB_DEF_SAD;        /* secondary GPIB address */
int mem_size = GPIB_DEF_MEM_SIZE;   /* memory (in kB) to reserve for driver */
int timo     = GPIB_DEF_TIMO;       /* default timeout */


/*-----------------------------------------------------------------------*/


Board_t board;
int gpib_status = 0;
int gpib_error  = 0;
unsigned long gpib_count  = 0;

#if defined GPIB_DEBUG
int PDEBUG_len = 0;
#endif


#ifdef CONFIG_PROC_FS
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 2, 0 )
static struct proc_dir_entry gpib_proc_entry = {
	0,
	5, "gpib0",
	S_IFREG | S_IRUGO,
	1, 0, 0,
	0,
	NULL,
	&gpib_read_proc
};
#endif
#endif


static bool gpib_board_struct_init( Board_t * board );


/*-----------------------------------------------------*
 * Function that gets called when the module is loaded
 *-----------------------------------------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
static int __init gpib_init( void )
#else
int init_module( void )
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
	int result;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) && \
    LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	struct proc_dir_entry *gpib_proc_entry;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 2, 0 )
	const char *namep;

	register_symtab( NULL );
#endif

	gpib_status = gpib_error = 0;

	/* Bring the board structure into a known initial state */

	printk( KERN_INFO "%s: Initializing module.\n", name );

	if ( ! gpib_board_struct_init( &board ) )
		return - EINVAL;

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
	if ( major == 0 )
		result = alloc_chrdev_region( &board.dev_no, 0, 1,
					      board.name );
	else {
		board.dev_no = MKDEV( major, minor );
		result = register_chrdev_region( board.dev_no, 1,
						 board.name );
	}

	if ( result < 0 ) {
		printk( KERN_ERR "%s: Can't register as char device.\n",
			board.name );
		goto device_registration_failure;
	}

	cdev_init( &board.ch_dev, &gpib_fops );
	board.ch_dev.owner = THIS_MODULE;

	if ( cdev_add( &board.ch_dev, board.dev_no, 1 ) < 0 ) {
		printk( KERN_ERR "%s: Can't register as char device.\n",
			board.name );
		unregister_chrdev_region( board.dev_no, 1 );
		goto device_registration_failure;
	}

	board.major = MAJOR( board.dev_no );

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 14 )
	board.gpib_class = class_create( THIS_MODULE, ( char * ) board.name );

	if ( IS_ERR( board.gpib_class ) ) {
		printk( KERN_ERR "%s: Can't create a class for "
			"the device.\n", board.name );
		cdev_del( &board.ch_dev );
		unregister_chrdev_region( board.dev_no, 1 );
		return -EIO;
	}

	class_device_create( board.gpib_class, NULL, major, NULL,
			     ( char * ) board.name );
#endif


#else
	/* Register the board as a character device with the major address
	   either compiled into the driver or set while the module is loaded.
	   If major is 0 the function returns a dynamically assigned major
	   device number, otherwise 0 on success (in which case we need to
	   set board.major to the correct value) */

	if ( ( board.major = register_chrdev( major, board.name,
					      &gpib_fops ) ) < 0 ) {
		printk( KERN_ERR "%s: Can't register as char device.\n", 
			board.name );
		goto device_registration_failure;
	}

	if ( major != 0 )
		board.major = major;
#endif

	/* Now boards may have to do hardware specific stuff, e.g. requesting
	   an I/O memory region. The function prints its own error messages. */

	if ( ! board_init( &board ) || board.type == 0 )
		goto board_init_failure;

	/* Initialize use of IRQ and DMA - this can fail due to invalid
	   settings for the IRQ or the DMA channel so that we have to give
	   up. If we can't get the interrupt or the DMA channel we try to
	   work without it. */

	if ( ! board_irq_dma_init( &board ) )
		goto irq_dma_init_failure;

	/* We haven't allocated memory yet */

	board.buf = NULL;

	/* Finally tell the world about the successful installation ;-) */

	printk( KERN_INFO "%s: Major = %d\n", board.name, board.major );

	printk( KERN_INFO "%s: Base = 0x%lx\n", board.name,
		( unsigned long ) board.base );

	if ( board.status & USE_IRQ )
		printk( KERN_INFO "%s: IRQ = %u\n", board.name, board.irq );
	else if ( board.use_irq )
		printk( KERN_INFO "%s: IRQ %u not used\n",
			board.name, board.irq );
	else
		printk( KERN_INFO "%s: IRQ disabled\n", board.name );

	if ( board.status & USE_DMA )
		printk( KERN_INFO "%s: DMA = %u\n", board.name, board.dma );
	else if ( board.use_dma )
		printk( KERN_INFO "%s: DMA %u not used\n",
			board.name, board.dma );
	else
		printk( KERN_INFO "%s: DMA disabled\n", board.name );

	printk( KERN_INFO "%s: Requested buffer size: %d kB\n",
		board.name, mem_size );

	board.total_read  = board.open_read = board.last_read =
	board.total_write = board.open_write = board.last_write = 0;
	board.last_error  = -1;
	board.gpib_status = gpib_status;

#ifdef CONFIG_PROC_FS
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
	create_proc_read_entry( board.name, 0, NULL, gpib_read_proc, NULL );
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) && \
    LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	if ( ( gpib_proc_entry = create_proc_entry( board.name, 0, NULL ) ) )
		gpib_proc_entry->read_proc = gpib_read_proc;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 2, 0 )
	for ( gpib_proc_entry.namelen = 0, namep = board.name;
	      *namep != '\0'; gpib_proc_entry.namelen++, namep++ )
		/* empty */ ;
	gpib_proc_entry.name = board.name;
	proc_register_dynamic( &proc_root, &gpib_proc_entry );
#endif
#endif

	return 0;

	/* Here follows the error handling... */

irq_dma_init_failure:
	board_cleanup( &board );

board_init_failure:
	unregister_chrdev( board.major, board.name );

device_registration_failure:
	printk( KERN_ERR "%s: Module installation failed.\n", board.name );
	return - EIO;
}


/*-------------------------------------------------------*
 * Function that gets called when the module is unloaded
 *-------------------------------------------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
static void __exit gpib_exit( void )
#else
void cleanup_module( void )
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 6, 0 )
	int ret;
#endif

	/* Get rid of the proc file */

#ifdef CONFIG_PROC_FS
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
	remove_proc_entry( board.name, NULL );
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) && \
    LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	remove_proc_entry( board.name, NULL );
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 2, 0 )
	proc_unregister( &proc_root, gpib_proc_entry.low_ino );
#endif
#endif
	/* Release interrupt and DMA channel */

	board_irq_dma_cleanup( &board );

	/* Release the memory region we've been using etc. */

	board_cleanup( &board );

	/* Deallocate transfer buffer memory (call function with size of 0) */

	if ( board.buf != NULL )
		gpib_get_memory( &board, 0 );

	/* Unregister the (char) device */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 14 )
	class_device_destroy( board.gpib_class, board.dev_no );
        class_destroy( board.gpib_class );
#endif

	unregister_chrdev_region( board.dev_no, 1 );
#else
	if ( ( ret = unregister_chrdev( board.major, board.name ) ) < 0 ) {
		printk( KERN_ERR "%s: Device busy or other module "
			"error: %d.\n", board.name, ret );
		return;
	}
#endif

	printk( KERN_INFO "%s: Module succesfully removed.\n",
		board.name );
}


/*--------------------------------------------------------*
 * Function sets up the board structure to a known state
 * and does a preliminary check of the boards parameters.
 *--------------------------------------------------------*/

static bool gpib_board_struct_init( Board_t * board )
{
	board->name = name == NULL ? "" : name;
	board->is_running = UNSET;

	if ( major < 0 || major > 255 ) {
		printk( KERN_ERR "%s: Invalid major device number: %d.\n",
			board->name, major );
		return FAIL;
	}

	if ( base < 0 ) {
		printk( KERN_ERR "%s: Invalid base address: %d.\n",
			board->name, base );
		return FAIL;
	}

	board->base  = ( void * ) base;

	/* IRQ and DMA handling: no DMA use without interrupt but IRQ without
	   DMA is ok */

	if ( irq < 0 ) {
		board->use_irq = UNSET;
		board->use_dma = UNSET;

		/* DMA can only be used with interrupts */

		if ( dma >= 0 )
			printk( KERN_INFO "DMA can't be used without "
				"interrupts, switching DMA off.\n" );
	} else {
		board->irq = ( unsigned int ) irq;
		board->use_irq = SET;

		if ( dma < 0 )
			board->use_dma  = UNSET;
		else {
			board->dma = ( unsigned int ) dma;
			board->use_dma = SET;
		}
	}

	board->status = 0;
	board->open_count = 0;
	board->type = 0;

	/* Set up the primary address(es) of the board - it's just the minor
	   device file number */

	if ( minor < 0 || minor > 0x1E ) {
		printk( KERN_ERR "%s: Invalid minor device number/GPIB "
			"address: %d.\n", board->name, minor );
		return FAIL;
	}
	board->pad = minor;

	/* If secondary address is not set to zero check that it's in the
	   range between 0x60 (96) and 0x7e (126) */

	if ( sad == 0 )
		board->sad = 0;
	else if ( sad < 0x60 || sad > 0x7e ) {
		printk( KERN_ERR "%s: Invalid secondary GPIB address: %d.\n",
			board->name, sad );
		return FAIL;
	} else
		board->sad = sad;

	/* Set up of timer stuff */

	if ( timo < TNONE || timo > T1000s ) {
		printk( KERN_ERR "%s: Invalid timeout value: %d.\n",
			board->name, timo );
		return FAIL;
	}

	board->timer_index = timo;
	board->timeout = UNSET;
	gpib_setup_timer( );

	if ( mem_size <= 0 || mem_size > 64 ) {
		printk( KERN_ERR "%s: Invalid memory size: %d.\n",
			board->name, mem_size );
		return FAIL;
	}

	/* Allocation of memory is done on call of open() */

	board->buf      = NULL;
	board->buf_size = 0;

	/* Initialize the lock semaphore, it must allow one call of down() */

	semaphore_init( &board->lock_sem, 1 );

	/* Default EOS byte is the Line Feed character and the eosmode is set
	   to do 8-bit comparison for the EOS byte, to send END with the last
	   byte or EOS and to stop reading on receiving the EOS byte */

	board->eos = '\n';
	board->eosmode = GPIB_BIN | GPIB_XEOS | GPIB_REOS;
	board_setup_eos_byte( board );

	return OK;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
module_param( name, charp, S_IRUGO );
module_param( base, int, S_IRUGO );
module_param( irq, int, S_IRUGO );
module_param( dma, int, S_IRUGO );
module_param( major, int, S_IRUGO );
module_param( minor, int, S_IRUGO );
module_param( sad, int, S_IRUGO );
module_param( mem_size, int, S_IRUGO );
module_param( timo, int, S_IRUGO );
#elif LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
MODULE_PARM( name, "s" );
MODULE_PARM_DESC( name, "Identifier string for module and device file name" );
MODULE_PARM( base, "i" );
MODULE_PARM_DESC( base, "Base address of GPIB board" );
MODULE_PARM( irq, "i" );
MODULE_PARM_DESC( irq, "IRQ level (-1 to disable use of interrupts)" );
MODULE_PARM( dma, "i" );
MODULE_PARM_DESC( dma, "DMA channel (-1 to disable use of DMA)" );
MODULE_PARM( major, "i" );
MODULE_PARM_DESC( major, "Major device number of GPIB board (1-254 or 0 "
		         "for dynamically assigned number)" );
MODULE_PARM( minor, "i" );
MODULE_PARM_DESC( minor, "Minor device number/primary GPIB address of board "
			 "(0-30)" );
MODULE_PARM( sad, "i" );
MODULE_PARM_DESC( sad, "Secondary GPIB address of board (0 for off or 96-126)" );
MODULE_PARM( mem_size, "i" );
MODULE_PARM_DESC( mem_size, "Buffer size (in kB) to be requested for GPIB "
		            "board (max. 64 kB)" );
MODULE_PARM( timo, "i" );
MODULE_PARM_DESC( timo, "Timeout setting (0-17)" );

EXPORT_NO_SYMBOLS;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
MODULE_AUTHOR( "Jens Thoms Toerring <jt@toerring.de>" );
MODULE_DESCRIPTION( "NI GPIB board (AT-GPIB/TNT4882, NI PCII(a)) driver" );
#endif

#if defined MODULE_LICENSE
MODULE_LICENSE( "GPL" );
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
module_init( gpib_init );
module_exit( gpib_exit );
#endif


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
