/*
  $Id: memory.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/

#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"



/* Maximum and minimum memory size for transfer buffer */

#define MEM_MAX_SIZE           64                     /* 64 kB */
#define MEM_MIN_SIZE           1                      /*  1 kB */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
#define MEM_QUANTUM            1024
#else
#define MEM_QUANTUM            1000
#endif


/*-----------------------------------------------------------------------*
 * This function trys to allocate kernel memory for the transfer buffer.
 * The size of the buffer has to be specified (where a zero or negative
 * size means just to deallocate the buffer!) in kB, and it will be re-
 * adjusted to a reasonable value if necessary. Returns true on success,
 * false on failure.
 *-----------------------------------------------------------------------*/

bool gpib_get_memory( Board_t * board,
		      int       size )
{
	void *buf = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
	gfp_t prio  = GFP_KERNEL;
#else
	int prio  = GFP_KERNEL;
#endif


	/* Get rid of all memory that already may have been allocated */

	if ( board->buf != NULL )
		kfree( board->buf );

	board->buf      = NULL;
	board->buf_size = 0;

	/* If the requested size is zero (or negative), i.e. we're  supposed
	   to free possibly previously allocated memory, we're already done */

	if ( size <= 0 )
		return OK;

	/* If necessary adjust the requested size to a reasonable value, i.e.
	   not larger than MEM_MAX_SIZE and not smaller than MEM_MIN_SIZE */

	if ( size > MEM_MAX_SIZE )
		size = MEM_MAX_SIZE;

	if ( size < MEM_MIN_SIZE )
		size = MEM_MIN_SIZE;

	/* If the board is supposed to use DMA we need to allocate memory
	   below MAX_DMA_ADDRESS, thus we have to use the GFP_DMA flag
	   with kmalloc() */

	if ( board->use_dma )
		prio |= GFP_DMA;

	/* Lets play it save - on errors we retry asking for a bit less, at
	   least until we hit the MEM_MIN_SIZE limit */

	while ( buf == NULL && size >= MEM_MIN_SIZE )
		if ( ( buf = kmalloc( size * MEM_QUANTUM, prio ) ) == NULL )
			size--;

	/* On success store restult in the boards structure */

	if ( buf != NULL ) {
		board->buf      = buf;
		board->buf_size = size * MEM_QUANTUM;
	}

	return buf != NULL;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
