/*
  $Id: proc_entry.c 687 2006-01-26 12:18:44Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/

#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"


#ifdef CONFIG_PROC_FS


/*-----------------------------------------------------------*
 * Gets called on read accesses to the proc filesystem entry
 * for the module and writes out lots of informations about
 * the boards state in human readable form.
 * The 'read count' and 'write count' entries are the number
 * of bytes read from or written to the bus since the module
 * was loaded, the board was opened and on the last read or
 * write operation. 'last error' is the type of the last
 * error detected and is reset on (re)opening the board.
 *-----------------------------------------------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
int gpib_read_proc( char *  buffer,
		    char ** start,
		    off_t   offset,
		    int     count,
		    int *   eof,
		    void *  data )
#else
int gpib_read_proc( char *  buffer,
		    char ** start,
		    off_t   offset,
		    int     count,
		    int     unused )
#endif
{
	char *buf = buffer;
	int max_count = count - 80;
	char *err_list[ ] = { "EDVR", "ECIC", "ENOL", "EADR", "EARG", "ESAC",
			      "EABO", "ENEB", "EDMA", "EBTO", "EOIP", "ECAP",
			      "EFSO", "EOWN", "EBUS", "ESTB", "ESRQ", "ECFG",
			      "EPAR", "ETAB", "ENSD", "ENWE", "ENTF", "EMEM" };
	char *time_list[ ] = { "infinite", "10us", "30us", "100us", "300us",
			       "1ms", "3ms", "10ms", "30ms", "100ms", "300ms",
			       "1s", "3s", "10s", "30s", "100s", "300s",
			       "1000s" };


	buf += sprintf( buf, "board type     : " );
	switch ( board.type )
	{
		case ATGPIB :
			buf += sprintf( buf,
				    "National Instruments AT-GPIB/TNT4882\n" );
			break;

		case PCII :
			buf += sprintf( buf, "National Instruments PCII\n" );
			break;

		case PCIIA :
		    buf += sprintf( buf, "National Instruments PCIIA\n" );
			break;

		default :
			buf += sprintf( buf, "unknown\n" );
			break;
	}

	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "module name    : %s\n", board.name );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "base address   : 0x%p\n", board.base );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	if ( board.use_irq )
		buf += sprintf( buf, "irq            : %u%s\n", board.irq,
			       board.status & USE_IRQ ? "" : " ( disabled )" );
	else
		buf += sprintf( buf, "irq            : ( disabled ) \n" );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	if ( board.use_dma )
		buf += sprintf( buf, "dma            : %u%s\n", board.dma,
			       board.status & USE_DMA ? "" : " ( disabled )" );
	else
		buf += sprintf( buf, "dma            : ( disabled ) \n" );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "buffer size    : %lu\n", board.buf_size );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "board major    : %d\n", board.major );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "prim. address  : %u\n", board.pad );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "sec. address   : %u\n", board.sad );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "timeout        : %u ( %s )\n", board.timer_index,
			time_list[ board.timer_index ] );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	if ( board.open_count )
		buf += sprintf( buf, "state          : opened (%d times)\n",
				board.open_count );
	else
		buf += sprintf( buf, "state          : closed\n" );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "eos char       : 0x%02x", board.eos );
	if ( board.eos >= ' ' && board.eos <= '~' )
		buf += sprintf( buf, " ( '%c' )", board.eos );
	else if ( board.eos == '\n' )
		buf += sprintf( buf, " ( linefeed )" );
	else if ( board.eos == '\r' )
		buf += sprintf( buf, " ( carriage return )" );
	buf += sprintf( buf, "\n" );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "eos mode       : 0x%02x ( %s %s %s %s )\n",
			board.eosmode,
			board.eosmode & GPIB_EOT  ? "EOT"  : "!EOT",
			board.eosmode & GPIB_BIN  ? "BIN"  : "!BIN",
			board.eosmode & GPIB_XEOS ? "XEOS" : "!XEOS",
			board.eosmode & GPIB_REOS ? "REOS" : "!REOS" );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "GPIB status    : 0x%02x", board.gpib_status );
	if ( board.gpib_status )
		buf += sprintf( buf, " ( " );
	if ( board.gpib_status & GPIB_ERR )
		buf += sprintf( buf, "ERR " );
	if ( board.gpib_status & GPIB_TIMO )
		buf += sprintf( buf, "TIMO " );
	if ( board.gpib_status & GPIB_END )
		buf += sprintf( buf, "END " );
	if ( board.gpib_status & GPIB_SRQI )
		buf += sprintf( buf, "SRQI " );
	if ( board.gpib_status & GPIB_RQS )
		buf += sprintf( buf, "RQS " );
	if ( board.gpib_status & GPIB_CMPL )
		buf += sprintf( buf, "CMPL " );
	if ( board.gpib_status & GPIB_CIC )
		buf += sprintf( buf, "CIC " );
	if ( board.gpib_status & GPIB_ATN )
		buf += sprintf( buf, "ATN " );
	if ( board.gpib_status & GPIB_TACS )
		buf += sprintf( buf, "TACS " );
	if ( board.gpib_status & GPIB_LACS )
		buf += sprintf( buf, "LACS " );
	if ( board.gpib_status & GPIB_DTAS )
		buf += sprintf( buf, "DTAS " );
	if ( board.gpib_status & GPIB_DCAS )
		buf += sprintf( buf, "DCAS " );
	buf += sprintf( buf, board.gpib_status ? ")\n" : "\n" );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "read count     : %lu %lu %lu\n", 
			board.total_read, board.open_read, board.last_read );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	buf += sprintf( buf, "write count    : %lu %lu %lu\n", 
			board.total_write, board.open_write, board.last_write );
	if ( ( int ) ( buf - buffer) >= max_count )
		goto proc_done;

	if ( board.last_error < 0 )
		buf += sprintf( buf, "last error     : none\n" );
	else
		buf += sprintf( buf, "last error     : %d ( %s )\n",
				board.last_error, 
				err_list[ board.last_error ] );

 proc_done:
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	*eof = 1;
#endif

	return ( int ) ( buf - buffer);
}


#endif /* CONFIG_PROC_FS */


/*
 * Overrides for Emacs so that the layout follows the Linux tabbing style.
 *
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
