/*
  $Id: util.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"


static int gpib_get_address( unsigned int    address,
			     unsigned char * pad,
			     unsigned char * sad );
static unsigned char do_ln( unsigned char pad,
			    unsigned char sad );



/*----------------------------------------------------------------*
 * This function must be called before any other command in order
 * to make the board the active controller.
 *----------------------------------------------------------------*/

int gpib_onl( int v )
{
	PDEBUGin( );

	board_get_state( &board );

	if ( v ) {
		/* Try to make board Active Controller */

		if ( ! ( gpib_status & GPIB_ATN ) &&
		     board_sic( &board ) & GPIB_ERR ) {
			gpib_status |= GPIB_ERR;
			gpib_error   = GPIB_ENEB;
			PDEBUGout( );
			return - EIO;
		}

		/* Raise the REN line */

		if ( ! ( board.status & REN ) &&
			 board_ren( &board, SET ) & GPIB_ERR ) {
			board_release_control( &board );
			gpib_status |= GPIB_ERR;
			gpib_error   = GPIB_ENEB;
			PDEBUGout( );
			return - EIO;
		}

		board.status |= REN;
	} else {
		/* Nothing to be done if board is not the controller */

		if ( ! ( gpib_status & GPIB_CIC ) )
			return 0;

		/* Reset REN line and bring board in idle controller state */

		board_ren( &board, UNSET );
		board_release_control( &board );
		gpib_status = 0;
	}

	PDEBUGout( );
	return 0;
}


/*-----------------------------------------------------*
 *-----------------------------------------------------*/

int gpib_wait( int flags )
{
	PDEBUGin( );

	gpib_status &= ~GPIB_TIMO;

	if ( flags & GPIB_TIMO )
		gpib_start_timer( );
	else
		board.timeout = UNSET;

	board_wait( &board, flags );

	if ( flags & GPIB_TIMO )
		gpib_stop_timer( );

	if ( board.timeout )
		gpib_status |= GPIB_TIMO;

	PDEBUGout( );

	return 0;
}


/*-----------------------------------------------------*
 *-----------------------------------------------------*/

int gpib_get_state( void )
{
	PDEBUGin( );
	board_get_state( &board );
	PDEBUGout( );
	return gpib_status;
}


/*-----------------------------------------------------*
 *-----------------------------------------------------*/

int gpib_ren( bool remote )
{
	int retval;


	PDEBUGin( );
	retval = board_ren( &board, remote );
	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}


/*-----------------------------------------------------*
 *-----------------------------------------------------*/

int gpib_gts( void )
{
	int retval;


	PDEBUGin( );
	retval = board_gts( &board );
	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}


/*-----------------------------------------------------*
 *-----------------------------------------------------*/

int gpib_sic( void )
{
	int retval;


	PDEBUGin( );
	retval = board_sic( &board );
	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}


/*-----------------------------------------------------*
 * Addresses the device with the address passed to the
 * function as listener and the board as talker.
 *-----------------------------------------------------*/

int gpib_setup_listener( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char buffer[ 6 ];
	unsigned long count = 0;


	PDEBUGin( );

	board_get_state( &board );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUGout( );
		return gpib_status;
	}

	/* The board being both talker and listener doesn't make sense */

	if ( pad == board.pad &&
	     ( ( ! sad && ! board.sad ) || sad == board.sad ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUGout( );
		return gpib_status;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( gpib_command_setup( ) < 0 ) {
		PDEBUGout( );
		return gpib_status;
	}

	/* Assemble the command string to make device listener and board
           talker */

	buffer[ count++ ] = UNT;
	buffer[ count++ ] = UNL;
	buffer[ count++ ] = pad | LAD;
	if ( sad )
		buffer[ count++ ] = sad;
	buffer[ count++ ] = board.pad | TAD;
	if ( board.sad )
		buffer[ count++ ] = board.sad;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	if ( board_cmd( &board ) & GPIB_ERR ) {
		board_get_state( &board );
		PDEBUGout( );
		return gpib_status;
	}

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return gpib_status;
}


/*-----------------------------------------------------*
 * Addresses the device with the address passed to the
 * function as talker and the board as listener.
 *-----------------------------------------------------*/

int gpib_setup_talker( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char buffer[ 6 ];
	unsigned long count = 0;


	PDEBUGin( );

	board_get_state( &board );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUGout( );
		return gpib_status;
	}

	/* The board being both talker and listener doesn't make sense */

	if ( pad == board.pad &&
	     ( ( ! sad && ! board.sad ) || sad == board.sad ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUGout( );
		return gpib_status;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( gpib_command_setup( ) < 0 ) {
		PDEBUGout( );
		return gpib_status;
	}

	/* Assemble the command string to make device talker and board listener
	   (we have to re-address the board as listener even if it's already
	   listener to make sure other listeners are un-addressed) */

	buffer[ count++ ] = UNT;
	buffer[ count++ ] = UNL;
	buffer[ count++ ] = board.pad | LAD;
	if ( board.sad )
		buffer[ count++ ] = board.sad;
	buffer[ count++ ] = pad | TAD;
	if ( sad )
		buffer[ count++ ] = sad;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	if ( board_cmd( &board ) & GPIB_ERR ) {
		board_get_state( &board );
		PDEBUGout( );
		return gpib_status;
	}

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return gpib_status;
}


/*-----------------------------------------------------*
 *-----------------------------------------------------*/

int gpib_ppc( unsigned int address,
	      int          state )
{
	unsigned char pad, sad;
	unsigned char buffer[ 6 ];
	unsigned int count = 0;
	int retval;


	PDEBUGin( );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EINVAL;
	}

	/* Check the second argument */

	if ( ( state & ~ 0x0f ) != PPD && ( state & ~ 0x0f ) != PPD ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUG( "Invalid PPC command: 0x%x\n", state );
		PDEBUGout( );
		return - EINVAL;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	/* Assemble the comand string parallel configure the device */

	buffer[ count++ ] = UNL;
	buffer[ count++ ] = pad | LAD;
	if ( sad )
		buffer[ count++ ] = sad;
	buffer[ count++ ] = PPC;
	buffer[ count++ ] = ( unsigned char ) ( state & 0xff );
	buffer[ count++ ] = UNL;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	retval = board_cmd( &board );

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}


/*-----------------------------------------------------*
 *-----------------------------------------------------*/

int gpib_rpp( void __user * ppr )
{
	int retval;
	char c;


	/* Check if we have write access to the user supplied character
	   pointer */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	if ( access_ok( VERIFY_WRITE, ppr, 1 ) == 0 ) {
#else
	if ( verify_area( VERIFY_WRITE, ppr, 1 ) != 0 ) {
#endif
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUGout( );
		return - EFAULT;
	}

	retval = board_rpp( &board, &c );

	if ( ! ( retval & GPIB_ERR ) ) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
		__copy_to_user( ppr, &c, 1 );
#else
		memcpy_tofs( ppr, &c, 1 );
#endif
	}

	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}


/*------------------------------------------*
 * Tries to determine if there's a listener
 *------------------------------------------*/

int gpib_ln( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char i;
	unsigned char ret = 0;


	PDEBUGin( );

	board_get_state( &board );

	/* Get device address and check it - we can't use the normally used
	   function gpib_get_address() here because the secondary address
	   may be 0xFF, indicating that all valid secondary addresses need
	   to be checked. */

	pad = ( unsigned char ) ( address & 0xff );
	sad = ( unsigned char ) ( ( address >> 8 ) & 0xff );

	if ( pad > 0x1e ||
	     ( sad != 0 && sad != 0xFF && ( sad < 0x60 || sad > 0x7e ) ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUG( "Invalid address %x -> pad = 0x%x, sad = 0x%x\n",
			address, pad, sad );
		PDEBUGout( );
		return 0;
	}

	if ( sad == 0xFF )            /* check all valid secondary addresses */
		for ( i = 0x60; i <= 0x7E && ! ret && 
			        ! ( gpib_status & GPIB_ERR ); i++ )
			ret = do_ln( pad, i );
	else
		ret = do_ln( pad, sad );

	/* Release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return ( int ) ret;
}


/*------------------------------------------*
 *------------------------------------------*/

static unsigned char do_ln( unsigned char pad,
			    unsigned char sad )
{
	unsigned int count = 0;
	unsigned char buffer[ 4 ];


	PDEBUGin( );

	/* Make sure board is CIC and the ATN line ist set */

	if ( gpib_command_setup( ) < 0 ) {
		PDEBUGout( );
		return 0;
	}

	buffer[ count++ ] = UNT;
	buffer[ count++ ] = UNL;
	buffer[ count++ ] = pad | LAD;
	if ( sad )
		buffer[ count++ ] = sad;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	/* Failing with no listener isn't an error in this case */

	if ( board_cmd( &board ) & GPIB_ERR )
	{
		if ( gpib_error & GPIB_ENOL )
		{
			board_gts( &board );
			gpib_status &= ~ GPIB_ERR;
			gpib_error   = 0;
		}
		return 0;
	}

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return sad != 0 ? sad : 1;
}


/*----------------------------------------------*
 * Sends a device trigger message to the device
 * with the address passed to the function.
 *----------------------------------------------*/

int gpib_trg( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char buffer[ 4 ];
	unsigned long count = 0;
	int retval;


	PDEBUGin( );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EINVAL;;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	/* Assemble the comand string to send the GET (group execute trigger)
	   message to the device */

	buffer[ count++ ] = UNL;
	buffer[ count++ ] = pad | LAD;
	if ( sad )
		buffer[ count++ ] = sad;
	buffer[ count++ ] = GET;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	retval = board_cmd( &board );

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}	


/*-------------------------------------------------*
 * Sends a selected device clear message to the
 * device with the address passed to the function.
 *-------------------------------------------------*/

int gpib_clr( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char buffer[ 4 ];
	unsigned long count = 0;
	int retval;


	PDEBUGin( );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EINVAL;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	/* Assemble the comand string to send the SDC (selected device clear)
	   message to the device */

	buffer[ count++ ] = UNL;
	buffer[ count++ ] = pad | LAD;
	if ( sad )
		buffer[ count++ ] = sad;
	buffer[ count++ ] = SDC;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	retval = board_cmd( &board );

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}	


/*------------------------------*
 * Sends a device clear message
 *------------------------------*/

int gpib_dcl( void )
{
	unsigned char buffer = DCL;
	int retval;


	PDEBUGin( );

	/* Make sure board is CIC and the ATN line ist set */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	/* Write out the command */

	board.transf.buf = &buffer;
	board.transf.cnt = 1;
	board.transf.is_last = SET;

	retval = board_cmd( &board );

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}	


/*--------------------------------------------------*
 * Brings a device into the local state (from REMS)
 *--------------------------------------------------*/

int gpib_local( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char buffer[ 6 ];
	unsigned long count = 0;
	int retval;


	PDEBUGin( );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUG( "Invalid address %x\n", address );
		PDEBUGout( );
		return - EINVAL;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	/* Assemble the comand string to send the GTO (goto local)
	   message to the device */

	buffer[ count++ ] = UNL;
	buffer[ count++ ] = pad | LAD;
	if ( sad )
		buffer[ count++ ] = sad;
	buffer[ count++ ] = GTL;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	retval = board_cmd( &board );

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}


/*----------------------------------------------*
 * Brings a device into the local lockout state
 *----------------------------------------------*/

int gpib_llo( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char buffer[ 6 ];
	unsigned long count = 0;
	int retval;


	PDEBUGin( );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUG( "Invalid address %x\n", address );
		PDEBUGout( );
		return - EINVAL;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	/* Assemble the comand string to send the LLO (local lockout)
	   message to the device */

	buffer[ count++ ] = UNL;
	buffer[ count++ ] = pad | LAD;
	if ( sad )
		buffer[ count++ ] = sad;
	buffer[ count++ ] = LLO;

	/* Write out the command string */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	retval = board_cmd( &board );

	/* Now finally release the ATN line */

	board_gts( &board );

	PDEBUGout( );
	return retval & GPIB_ERR ? - EIO : 0;
}


/*------------------------------------------------------------------*
 * Function tries to make the board active controller, the argument
 * determines if this happens asynchronously (if argument is zero)
 * or synchronously (for other values of the argument).
 *------------------------------------------------------------------*/

int gpib_cac( int how )
{
	PDEBUGin( );

	if ( ! ( board_get_state( &board ) & GPIB_CIC ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return - EIO;
	}

	if ( how == 0 )
		board_tca( &board );
	else
		board_tcs( &board );

	PDEBUGout( );
	return gpib_status & GPIB_ERR ? - EIO : 0;
}


/*--------------------------------------------*
 *--------------------------------------------*/

int gpib_rsp( unsigned int address )
{
	unsigned char pad, sad;
	unsigned char buffer[ 8 ];
	unsigned long count = 0;
	int retval;
	unsigned char stb;


	PDEBUGin( );

	/* Get device address and check it */

	if ( gpib_get_address( address, &pad, &sad ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EINVAL;
	}

	/* Make sure board is CIC and the ATN line ist set */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	/* Assemble command to send SPE (serial poll enable) message
	   to device */

	buffer[ count++ ] = UNL;
	buffer[ count++ ] = board.pad | LAD;
	if ( board.sad )
		buffer[ count++ ] = board.sad;
	buffer[ count++ ] = UNT;
	buffer[ count++ ] = pad | TAD;
	if ( sad )
		buffer[ count++ ] = sad;
	buffer[ count++ ] = SPE;

	/* Send the command, return if not complete */

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	if ( ! ( board_cmd( &board ) & GPIB_CMPL ) ) {
		board_gts( &board );
		PDEBUGout( );
		return - EIO;
	}
	gpib_status &= ~ GPIB_CMPL;

	/* Now read in the status byte from the device (after un-asserting the
	   ATN line) */

	if ( board_gts( &board ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EIO;
	}

	board.transf.buf     = &stb;
	board.transf.cnt     = 1;
	board.transf.is_last = SET;
	board.transf.dir     = DMA_RD;

	gpib_start_timer( );
	retval = board_read( &board );
	gpib_stop_timer( );

	/* We still need to send the SPD (serial poll disable) message,
	   first set the ATN line again */

	if ( ( retval = gpib_command_setup( ) ) < 0 ) {
		PDEBUGout( );
		return retval;
	}

	count = 0;
	buffer[ count++ ] = SPD;
	buffer[ count++ ] = UNT;

	board.transf.buf = buffer;
	board.transf.cnt = count;
	board.transf.is_last = SET;

	board_cmd( &board );

	if ( board_gts( &board ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EIO;
	}

	/* Return in gpib_count the byte returned by serial poll */

	gpib_count = 0xFFUL & ( unsigned long ) stb;
	PDEBUGout( );
	return ( retval | gpib_status ) & GPIB_ERR ? - EIO : 0;
}


/*--------------------------------------------*
 *--------------------------------------------*/

static int gpib_get_address( unsigned int    address,
			     unsigned char * pad,
			     unsigned char * sad )
{
	PDEBUGin( );

	*pad = ( unsigned char ) ( address & 0xff );
	*sad = ( unsigned char ) ( ( address >> 8 ) & 0xff );

	if ( *pad > 0x1e ||
	     ( *sad != 0 && ( *sad < 0x60 || *sad > 0x7e ) ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUG( "Invalid address %x -> pad = 0x%x, sad = 0x%x\n",
			address, *pad, *sad );
	}

	PDEBUGout( );
	return gpib_status;
}


/*---------------------------------------------------------------------*
 * Function sets the EOS character as well as the eosmode. Both values
 * have to be stored in the board structure and set in the correct
 * registers. We have to distinguish between eosmode, that only states
 * the mode and eos_byte which is the value actually written to the
 * boards register. To calculate the latter we have to call the
 * function board_setup_eos_byte().
 *---------------------------------------------------------------------*/

int gpib_eos( int v )
{
	unsigned char new_eos;
	unsigned int new_eosmode;


	PDEBUGin( );

	new_eos = ( unsigned char ) ( v & 0xff );
	new_eosmode = ( unsigned int ) ( ( v >> 8 ) & 0xff );


	if ( new_eos != board.eos ) {
		board.eos = new_eos;
		board_set_eos( &board );
	}

	if ( new_eosmode != board.eosmode ) {
		board.eosmode = new_eosmode;
		board_setup_eos_byte( &board );
		board_set_eosmode( &board, 0 );
	}

	PDEBUGout( );
	return 0;
}


/*---------------------------------------------------*
 * Function sets a new primary address of the board.
 *---------------------------------------------------*/

int gpib_pad( int pad )
{
	PDEBUGin( );
	if ( pad < 0 || pad > 0x1e )
		return - EINVAL;

	board.pad = pad;
	board_pad( &board );

	board_get_state( &board );

	PDEBUGout( );
	return 0;
}


/*-------------------------------------------------------------*
 * Function sets a new secondary address of the board - if the
 * secondary is 0 secondary addressing gets switched off.
 *-------------------------------------------------------------*/

int gpib_sad( int sad )
{
	PDEBUGin( );
	if ( sad != 0 && ( sad < 0x60 || sad > 0x7e ) )
		return - EINVAL;

	board.sad = sad;
	board_sad( &board );

	board_get_state( &board );

	PDEBUGout( );
	return 0;
}


/*-------------------------------------------------------------------------*
 * Brings board in active controller state as needed for sending commands.
 * Obviously, this can't work if the board isn't controller at all.
 *-------------------------------------------------------------------------*/

int gpib_command_setup( void )
{
	PDEBUGin( );

	/* The board needs to be at least CIC to continue */

	if ( ! ( board_get_state( &board ) & GPIB_CIC ) ) {
		gpib_error = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return - EIO;
	}

	/* We need to become active controller to send commands - take care:
	   if the board is the primary addressed talker we've got to take
	   control asynchronously, otherwise synchrounously will do, at least
	   the AT-GPIB/TNT board with the TNT4882 chip needs it this way */

	if ( ! ( gpib_status & GPIB_ATN ) ) {
		if ( gpib_status & GPIB_TACS )
			board_tca( &board );
		else
			board_tcs( &board );
	}

	PDEBUGout( );
	return gpib_status & GPIB_ERR ? - EIO : 0;
}


/*------------------------------------------------------------------*
 * Returns the state of the GPIB's control lines in the upper byte
 * of 'gpib_count' while corresponding bits the lower byte indicate
 * if the  returned values are values (this is necessary because
 * some boards can't test the states of the control lines).
 *------------------------------------------------------------------*/

int gpib_lines( void )
{
	int ret;


	PDEBUGin( );
	board_get_state( &board );          /* must be called first ! */
	ret = board_lines( &board );
	PDEBUGout( );
	return ret;
}


/*-------------------------------------------------------------------*
 * Sets a semaphore count to a certain value - this function is here
 * to avoid the hassle of having lots of defines all over the place
 * just because of the handling in different kernel versions...
 *-------------------------------------------------------------------*/

void semaphore_init( struct semaphore * semaphore,
		     int                state )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	sema_init( semaphore, state );
#else
	semaphore->count = state;
#endif
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
