/*
  $Id: board.c 736 2006-07-07 15:48:51Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "lib.h"


int gpib_wait( int dev_dscr,
               int mask );
int gpib_sre( int v );
int gpib_gts( void );
int gpib_dcl( void );
int gpib_cac( int v );
int gpib_sic( void );
int gpib_ln( int         pad,
             int         sad,
             short int * listener );
int gpib_irq( int state );
int gpib_dma( int state );
int gpib_rpp( char * ppr );
int gpib_lines( int * lines );

extern int gpib_ioctl( int cmd,
                       ... );
extern Device *check_dscr( int dev_dscr );
extern int gpib_setup_timo( Device * dev );


extern Device devices[ ];
extern int num_devices;
extern int master_dev;

extern int gpib_status;
extern int gpib_error;
extern unsigned long gpib_count;

extern bool board_is_online;


/*---------------------------------------*
 *---------------------------------------*/

int gpib_wait( int dev_dscr,
               int mask )
{
        Device *dev;


        /* Check that the device descriptor */

        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Only waits for ERR, TIMO, END and RQS condition are accepted */

        if ( mask & ~ ( GPIB_ERR | GPIB_TIMO | GPIB_END | GPIB_RQS ) )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Calls with an empty mask or for ERR only return immediately */

        if ( mask == 0 || mask == GPIB_ERR )
                return gpib_status;

        /* Set timeout value for the device the function was called for */

        if ( mask & GPIB_TIMO && gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        gpib_ioctl( GPIB_IOC_WAIT, mask );
        return gpib_status;
}


/*----------------------------------------------------------*
 *----------------------------------------------------------*/

int gpib_sre( int v )
{
        /* Set or reset the REN line */

        return gpib_ioctl( GPIB_IOC_REN, v );
}


/*----------------------------------------------------------*
 *----------------------------------------------------------*/

int gpib_gts( void )
{
        /* Unassert the ATN line */

        return gpib_ioctl( GPIB_IOC_GTS );
}


/*----------------------------------------------------------*
 *----------------------------------------------------------*/

int gpib_dcl( void )
{
        /* Send the DCL message */

        return gpib_ioctl( GPIB_IOC_DCL );
}


/*----------------------------------------------------------*
 *----------------------------------------------------------*/

int gpib_cac( int v )
{
        /* Try to assert the ATN line, either synchronously (if 'v' is
           non-zero) or asynchronously (if 'v' is zero) */

        return gpib_ioctl( GPIB_IOC_CAC, v );
}


/*----------------------------------------------------------*
 *----------------------------------------------------------*/

int gpib_sic( void )
{
        /* Function raises the IFC line for at leat 100 us */

        return gpib_ioctl( GPIB_IOC_SIC );
}       


/*----------------------------------------------------------*
 *----------------------------------------------------------*/

int gpib_ln( int         pad,
             int         sad,
             short int * listener )
{
        /* Check the parameter */

        if ( pad < 0 || pad > 0x1e ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        if ( sad != NO_SAD && sad != ALL_SAD &&
             ( sad < 0x60 || sad > 0x7E ) ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        if ( pad == devices[ master_dev ].pad &&
             ( sad == devices[ master_dev ].sad || sad == ALL_SAD ) ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        *listener = ( short int ) gpib_ioctl( GPIB_IOC_LN,
                                              ADDRESS( pad, sad ) );
        return gpib_status;
}


/*-------------------------------------------------------------------------*
 * Function switches the use of interrupts on or off (of course, interrupts
 * can only be switched on if the driver was loaded with options allowing
 * tthe use of interrupts).
 *--------------------------------------------------------------------------*/

int gpib_irq( int state )
{
        return gpib_ioctl( GPIB_IOC_IRQ, state );
}


/*--------------------------------------------------------------------*
 * Function switches use of DMA on or off (of course, DMA can only be
 * switched on if the driver was loaded with options allowing use of
 * DMA).
 *--------------------------------------------------------------------*/

int gpib_dma( int state )
{
        return gpib_ioctl( GPIB_IOC_DMA, state );
}


/*----------------------------------------------------------*
 *----------------------------------------------------------*/

int gpib_rpp( char * ppr )
{
        if ( ppr == NULL ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        return gpib_ioctl( GPIB_IOC_RPP );
}


/*---------------------------------------------------------------------------*
 * Function returns the current states of the control lines of the GPIB bus.
 * The bits in the lower byte indicate if the corresponding line can be 
 * monitored while the upper byte contains the bits indicating the states of
 * the lines, see also definitions of ValidXXX and BusXXX in gpib.h.
 *---------------------------------------------------------------------------*/

int gpib_lines( int * lines )
{
        *lines = gpib_ioctl( GPIB_IOC_LINES, lines );
        return gpib_status;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
