/*
  $Id: control.c 736 2006-07-07 15:48:51Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "lib.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


#define GPIB_LIB_MAIN

Device devices[ GPIB_MAX_DEVICES ];
Device devices_orig[ GPIB_MAX_DEVICES ];
int num_devices = 0;
int master_dev = -1;
int master_file;

static const char *cfg;

int gpib_status;
int gpib_error;
unsigned long gpib_count;

char *gpib_error_buffer = NULL;
size_t gpib_error_buffer_length = 0;

bool board_is_online = UNSET;


int gpib_find( const char * dev_name );
int gpib_dev( int pad,
              int sad,
              int timo,
              int eos_char,
              int eos_mode );
int gpib_ioctl( unsigned int cmd,
                ... );

int gpib_find_device( const char * dev_name );
static int gpib_parse_config( const char * file );
Device * gpib_find_master( void );
static int gpib_setup_master( void );
int gpib_dev_setup( Device * temp_dev );


extern int yyparse( void );
extern Device * check_dscr( int dev_dscr );
extern int dev_line;



/*------------------------------------------------------------------*
 * This function allows the calling program to define a buffer to
 * which error messages are written. The maximum length of an error
 * message is set by the second parameter.
 *------------------------------------------------------------------*/

void gpib_error_setup( char * buffer,
                       size_t buffer_length )
{
        gpib_error_buffer = buffer;
        gpib_error_buffer_length = buffer_length;
}


/*-----------------------------------------------------------------*
 * This function is called to initialize a device and find out
 * from the device name the number for the device to be used in
 * further function calls.
 * Usually, this function will bring the device for which it has
 * been called in the remote state, locking the keyboard.
 * On the very first call (or when the board is not online, i.e.
 * not working as GPIB bus master) the configuration file is being
 * read in and the board is brought online.
 *-----------------------------------------------------------------*/

int gpib_find( const char * dev_name )
{
        int dev_dscr = 0;
        Device *dev;
        int i;


        /* Get current value of gpib_state */

        if ( board_is_online )
                gpib_ioctl( GPIB_IOC_STATE );
        else {
                gpib_status &= ~ ( GPIB_CMPL | GPIB_ERR |
                                   GPIB_END | GPIB_TIMO );
                gpib_error   = 0;
        }

        /* Check that the device name is not completely bogus - except when the
           board isn't already online a missing device name is an error. Only
           then an empty device name is treated as a call for the master
           device. */

        if ( ( dev_name == NULL || dev_name[ 0 ] == '\0' ) &&
             board_is_online ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;

                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Empty device name in call of "
                                  "gpib_find()." );

                return -1;
        }
        
        /* If this is the very first call (or when the board isn't online) read
           and parse the configuration file and bring the board online. Use
           either the environment variable GPIB_CONF or the built-in name of
           the configuration file (environment variable takes precedence). */

        if ( ! board_is_online ) {
                for ( i = 0; i < GPIB_MAX_DEVICES; i++ )
                {
                        *devices[ i ].name  = '\0';
                        devices[ i ].flags  = 0;
                        *devices[ i ].file  = '\0';
                        *devices[ i ].ename = '\0';
                }

                cfg = getenv( "GPIB_CONF" );

                if ( cfg == NULL )
                        cfg = GPIB_DEF_CONFIG_FILE;

                if ( gpib_parse_config( cfg  ) != 0 ) {
                        master_dev = -1;
                        num_devices = 0;
                        gpib_status |= GPIB_ERR;
                        return -1;
                }

                if ( ( dev_dscr = gpib_setup_master( ) ) != 0 )
                {
                        master_dev = -1;
                        num_devices = 0;
                        return -1;
                }

                board_is_online = SET;
        }

        /* If this isn't a call for the master device (indicated by an empty
           device file name) try to find the device in the list - if not found
           return -1 (gpib_status and gpib_error are already set correctly) */

        if ( dev_name != NULL && dev_name[ 0 ] != '\0' &&
             ( dev_dscr = gpib_find_device( dev_name ) ) == -1 )
                return -1;

        dev = &devices[ dev_dscr ];

        if ( ! ( dev->flags & IS_INIT ) && ! ( dev->flags & IS_MASTER ) ) {

                /* Check that the address is unique */

                for ( i = 0; i < num_devices; i++ )
                        if ( devices + i != dev &&
                             devices[ i ].flags & IS_INIT &&
                             ADDRESS( devices[ i ].pad, devices[ i ].sad ) ==
                             ADDRESS( dev->pad, dev->sad ) ) {
                                gpib_status |= GPIB_ERR;
                                gpib_error   = GPIB_EARG;

                                if ( gpib_error_buffer != NULL &&
                                     gpib_error_buffer_length > 0 )
                                        snprintf( gpib_error_buffer,
                                                  gpib_error_buffer_length,
                                                  "Device \"%s\" and \"%s\" "
                                                  "have the same primary and "
                                                  "secondary GPIB address.",
                                                  devices[ i ].name,
                                                  dev->name );
                                return -1;
                        }

                dev->flags |= IS_INIT;
        }

        return dev_dscr;
}


/*-------------------------------------------------------------*
 * This is an alternative method to get at a device descriptor
 * for devices that aren't listed in the configuration file.
 *-------------------------------------------------------------*/

int gpib_dev( int pad,
              int sad,
              int timo,
              int eos_char,
              int eos_mode )
{
        int i;
        int dev_dscr = 0;


        if ( board_is_online )
                gpib_ioctl( GPIB_IOC_STATE );
        else {
                gpib_status &= ~ ( GPIB_CMPL | GPIB_ERR |
                                   GPIB_END | GPIB_TIMO );
                gpib_error   = 0;
        }

        /* We can't have more than GPIB_MAX_DEVICES */

        if ( num_devices >= GPIB_MAX_DEVICES ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Internal table overflow, too many "
                                  "devices." );
                gpib_status |= GPIB_ETAB;
                gpib_error   = GPIB_EARG;
                return -1;
        }

        /* Check the paramter */

        if ( pad < 0 || pad >= 0x1e ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid primary GPIB address %d.", pad );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return -1;
        }

        if ( sad != NO_SAD && (  sad < 0x60 || sad > 0x7e ) ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid secondary GPIB address %d", sad );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return -1;
        }

        if ( timo < TNONE || timo > T1000s ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid timeout setting: %d.", timo );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return -1;
        }

        if ( eos_char < 0 || eos_char > 0xFF ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid EOS character." );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return -1;
        }

        if ( eos_mode < 0 || eos_mode > 0xF ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid EOS mode." );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return -1;
        }

        /* If the board hasn't been broad online it's now time to do so */

        if ( ! board_is_online && gpib_find( NULL ) == -1 )
        {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Failed to bring board online." );
                return -1;
        }

        /* Check that the devices addresses do not clash with the ones of an
           already initiaized device */

        for ( i = 0; i < num_devices; i++ )
                if ( devices[ i ].flags & IS_INIT &&
                     ADDRESS( devices[ i ].pad, devices[ i ].sad ) ==
                     ADDRESS( pad, sad ) ) {
                        if ( gpib_error_buffer != NULL &&
                             gpib_error_buffer_length > 0 )
                                snprintf( gpib_error_buffer,
                                          gpib_error_buffer_length,
                                          "Device has the same primary and "
                                          "secondary GPIB address as device "
                                          "\"%s\".", devices[ i ].name );
                        gpib_status |= GPIB_ERR;
                        gpib_error   = GPIB_EARG;
                        return -1;
                }

        dev_dscr = num_devices++;

        sprintf( devices[ dev_dscr ].name, "unnamed_%d_%d", pad, sad );
        devices[ dev_dscr ].pad     = pad;
        devices[ dev_dscr ].sad     = sad;
        devices[ dev_dscr ].eos     = eos_char;
        devices[ dev_dscr ].eosmode = eos_mode;
        devices[ dev_dscr ].timo    = timo;

        /* Store states of devices as just read in */

        memcpy( devices_orig + dev_dscr, devices + dev_dscr,
                sizeof( Device ) );

        devices[ dev_dscr ].flags = IS_INIT;
        
        return dev_dscr;
}


/*------------------------------------------------------------------*
 * Function returns the internal number of the device with the name
 * passed to the function or -1 if no device with this name could
 * be found. The function also so sets gpib_status and gpib_error
 * on errors.
 *------------------------------------------------------------------*/

int gpib_find_device( const char * dev_name )
{
        int i;


        for ( i = 0; i < num_devices; i++ )
                if ( ! strcmp( dev_name, devices[ i ].name ) )
                        break;

        if ( i >= num_devices ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Failed to find device \"%s\" in "
                                  "configuration file \"%s\".",
                                  dev_name, cfg );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EDVR;
                return -1;
        }

        return i;
}


/*--------------------------------------------------------*
 * Function opens the configuration file, runs the parser
 * and finally checks the plausibility of the results.
 *--------------------------------------------------------*/

static int gpib_parse_config( const char * file )
{
        extern FILE *yyin;
        FILE *fp;
        int retval;


        if ( ( fp = fopen( file, "r" ) ) == NULL ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_ECFG;
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Failed to open configuration file \"%s\".",
                                  cfg );
                return -1;
        }

        yyin = fp;
        if ( ( retval = yyparse( ) ) != 0 )
        {
                gpib_status |= GPIB_ERR;
                gpib_error = GPIB_EPAR;
        }

        fclose( fp );

        return retval;
}


/*--------------------------------------------------------------*
 *--------------------------------------------------------------*/

Device * gpib_find_master( void )
{
        int i;
        Device tmp_dev;


        for ( i = 0; i < num_devices; i++ )
                if ( devices[ i ].flags & IS_MASTER )
                        break;

        if ( i == num_devices ) {            /* no master device ? */
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "No master device found in configuration "
                                  "file \"%s\".", cfg );
                gpib_error  = GPIB_ENEB;
                gpib_status = GPIB_ERR;
                return NULL;
        }

        master_dev = i;

        /* Bring the master device to the top of the list so that it is
           guaranteed to have the device descriptor 0 */

        memcpy( &tmp_dev, devices + master_dev, sizeof *devices );
        memcpy( devices, devices + master_dev, sizeof *devices );
        memcpy( devices, &tmp_dev, sizeof *devices );

        /* Finally check that there isn't more than one master device */

        for ( i = 1; i < num_devices; i++ )
                if ( devices[ i ].flags & IS_MASTER ) {
                        if ( gpib_error_buffer != NULL &&
                             gpib_error_buffer_length > 0 )
                                snprintf( gpib_error_buffer,
                                          gpib_error_buffer_length,
                                          "More than one master device found "
                                          "in configuration file \"%s\".",
                                          cfg );
                        gpib_error  = GPIB_ECFG;
                        gpib_status = GPIB_ERR;
                        return NULL;
                }

        /* Store states of devices as of now */

        memcpy( devices_orig, devices, num_devices * sizeof( Device ) );

        return devices;
}


/*------------------------------------------------------------*
 * Function checks if there is a device designated as master,
 * opens the corresponding device file, sets the primary (and
 * if set the secondary) address and brings the board online.
 *------------------------------------------------------------*/

static int gpib_setup_master( void )
{
        Device *master;


        if ( ( master = gpib_find_master( ) ) == NULL )
                return -1;

        /* Assemble device file name of master device by appending the device
           name to the default path if it hasn't been set in the configuration
           file */

        if ( master->file == NULL || *master->file == '\0' ) {
                strcpy( master->file, GPIB_DEF_DEVICE_DIR );
                if ( master->file[ strlen( master->file ) - 1 ] != '/' )
                        strcat( master->file, "/" );
                strcat( master->file, master->name );
        }

        /* Try to open the device file (don't wait if someone else has it
           already open) - this also brings the board online */

        if ( ( master_file =
               open( master->file, O_RDWR | O_NONBLOCK ) ) < 0 ) {

                gpib_status |= GPIB_ERR;

                if ( errno == EBUSY ) {
                        if ( gpib_error_buffer != NULL &&
                             gpib_error_buffer_length > 0 )
                                snprintf( gpib_error_buffer,
                                          gpib_error_buffer_length,
                                          "Master device file \"%s\" locked "
                                          "by another process.",
                                          master->file );
                        gpib_error = GPIB_ELCK;
                }
                else if ( errno == ENXIO || errno == ENODEV )
                {
                        if ( gpib_error_buffer != NULL &&
                             gpib_error_buffer_length > 0 )
                                snprintf( gpib_error_buffer,
                                          gpib_error_buffer_length,
                                          "GPIB device module not loaded." );
                        gpib_error = GPIB_EDVR;
                } else {
                        if ( gpib_error_buffer != NULL &&
                             gpib_error_buffer_length > 0 )
                                snprintf( gpib_error_buffer,
                                          gpib_error_buffer_length,
                                          "Failed to open master device "
                                          "file \"%s\".", master->file );
                        gpib_error = GPIB_ENEB;
                }

                return -1;
        }

        /* Set the FD_CLOEXEC bit for the device file */

        fcntl( master_file, F_SETFD, FD_CLOEXEC );

        /* Set the primary (and, if set, the secondary) address */

        if ( gpib_ioctl( GPIB_IOC_PAD, master->pad ) & GPIB_ERR ||
             ( master->sad &&
               gpib_ioctl( GPIB_IOC_SAD, master->sad ) & GPIB_ERR ) ) {
                close( master_file );
                gpib_error   = GPIB_ENEB;
                gpib_status |= GPIB_ERR;
                return -1;
        }

        /* Everything seems to have worked out fine... */

        master->flags |= IS_INIT;

        return 0;
}


/*---------------------------------------------------------------------*
 * Function is called from the parser to check a few device parameters
 * and to copy them to the appropriate device structure.
 *---------------------------------------------------------------------*/

int gpib_dev_setup( Device * temp_dev )
{
        int i;


        /* Take care that there aren't too many devices */

        if ( num_devices >= GPIB_MAX_DEVICES )
                return -1;

        /* Check validity of name, addresses and timeout flag */

        if ( *temp_dev->name == '\0' ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Name of device missing near line %d in "
                                  "\"%s\".", dev_line, cfg );
                return -1;
        }

        if ( temp_dev->pad < 0 || temp_dev->pad > 0x1e ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid primary GPIB address %d for device "
                                  "\"%s\".", temp_dev->pad, temp_dev->name );
                return -1;
        }

        if ( temp_dev->sad != NO_SAD &&
                 ( temp_dev->sad < 0x60 || temp_dev->sad > 0x7e ) ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid secondary GPIB address %d for "
                                  "device \"%s\".",
                                  temp_dev->sad, temp_dev->name );
                return -1;
        }

        if ( temp_dev->timo < TNONE || temp_dev->timo > T1000s ) {
                if ( gpib_error_buffer != NULL &&
                     gpib_error_buffer_length > 0 )
                        snprintf( gpib_error_buffer, gpib_error_buffer_length,
                                  "Invalid timeout setting for device \"%s\".",
                                  temp_dev->name );
                return -1;
        }

        /* Check that the device name is unique */

        for ( i = 0; i < num_devices; i++ )
                if ( ! strcmp( devices[ i ].name, temp_dev->name ) )
                {
                        if ( gpib_error_buffer != NULL &&
                             gpib_error_buffer_length > 0 )
                                snprintf( gpib_error_buffer,
                                          gpib_error_buffer_length,
                                          "Device \"%s\" found more than once "
                                          "in configuration file \"%s\".",
                                          temp_dev->name, cfg );
                        return -1;
                }

        /* Copy to real device structure and increment device count */

        memcpy( devices + num_devices++, temp_dev, sizeof( Device ) );

        return 0;
}


/*---------------------------------------------------------*
 * Here's the place where the interesting stuff happens...
 *---------------------------------------------------------*/

int gpib_ioctl( unsigned int cmd,
                ... )
{
        va_list ap;
        GPIB_Arg gpib_arg;
        size_t cnt;
        

        switch ( cmd ) {
                case GPIB_IOC_READ  :     /* read data from a device         */
                case GPIB_IOC_WRT   :     /* send data to a device           */
                case GPIB_IOC_DVRD  :
                case GPIB_IOC_DVWRT :
                        va_start( ap, cmd );
                        gpib_arg.arg = va_arg( ap, int );
                        gpib_arg.buf = ( void * ) va_arg( ap, char * );
                        cnt = va_arg( ap, size_t );
                        va_end( ap );
                        gpib_arg.cnt = ( unsigned long ) cnt;
                        break;

                case GPIB_IOC_CMD   :     /* write commands to bus           */
                        va_start( ap, cmd );
                        gpib_arg.buf = ( void * ) va_arg( ap, char * );
                        cnt = va_arg( ap, size_t );
                        va_end( ap );
                        gpib_arg.cnt = ( unsigned long ) cnt;
                        break;

                case GPIB_IOC_WAIT  :     /* wait for event                  */
                case GPIB_IOC_ONL   :     /* bring board on- or offline      */
                case GPIB_IOC_REN   :     /* raise/lower REN line            */
                case GPIB_IOC_TIMO  :     /* set new timeout value           */
                case GPIB_IOC_IRQ   :     /* switch on/off use of interrupt  */
                case GPIB_IOC_DMA   :     /* switch on/off use of DMA        */
                case GPIB_IOC_TRG   :     /* trigger a device                */
                case GPIB_IOC_CAC   :     /* raise the ATN line              */
                case GPIB_IOC_SDC   :     /* clear a device                  */
                case GPIB_IOC_LOCAL :     /* unlock a device                 */
                case GPIB_IOC_PAD   :     /* set primary address of master   */
                case GPIB_IOC_SAD   :     /* set secondary address of master */
                case GPIB_IOC_EOS   :     /* set new EOS / eosmode           */
                case GPIB_IOC_RSP   :     /* serial poll a device            */
                case GPIB_IOC_LN :        /* find listener                   */
                        va_start( ap, cmd );
                        gpib_arg.arg = va_arg( ap, int );
                        va_end( ap );
                        break;

                case GPIB_IOC_PPC :       /* parallel poll configure         */
                        va_start( ap, cmd );
                        gpib_arg.arg  = va_arg( ap, int );
                        gpib_arg.stat = va_arg( ap, int );
                        va_end( ap );
                        break;

                case GPIB_IOC_RPP :       /* conduct parallel poll           */
                        va_start( ap, cmd );
                        gpib_arg.buf = va_arg( ap, char * );
                        va_end( ap );
                        break;

                case GPIB_IOC_SIC   :     /* raise IFC and ATN line          */
                case GPIB_IOC_GTS   :     /* release ATN line                */
                case GPIB_IOC_DCL   :     /* send device clear command       */
                case GPIB_IOC_STATE :     /* get state in gpib_status        */
                case GPIB_IOC_LINES :     /* get lines in gpib_count         */
                        break;

                default :
                        gpib_status = GPIB_ERR;
                        gpib_error  = GPIB_EARG;
                        return gpib_status;
        }

        ioctl( master_file, cmd, &gpib_arg );

        gpib_status = gpib_arg.stat;
        gpib_error  = gpib_arg.err;
        gpib_count  = gpib_arg.cnt;

        if ( cmd == GPIB_IOC_IRQ || cmd == GPIB_IOC_DMA ||
             cmd == GPIB_IOC_LINES || cmd == GPIB_IOC_LN ||
             cmd == GPIB_IOC_RSP )
                return gpib_arg.arg;

        return gpib_status;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
