/*
  $Id: io.c 736 2006-07-07 15:48:51Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "lib.h"


extern Device devices[ ];
extern int num_devices;
extern int master_dev;

extern int gpib_status;
extern int gpib_error;
extern unsigned long gpib_count;

extern bool board_is_online;

extern int gpib_ioctl( int cmd,
                       ... );
extern int gpib_setup_timo( Device * dev );
extern int gpib_setup_eos( Device * dev );
extern Device *check_dscr( int dev_dscr );



/*------------------------------------------------------------*
 * Writes 'count' bytes from 'buffer' to the device addressed
 * by 'dev_dscr' (which was returned by a previous call of
 * gpib_find()). Before the data are send to the device the
 * EOS character and EOS mode for the device are set as well
 * as the timeout value.
 *------------------------------------------------------------*/

int gpib_wrt( int          dev_dscr,
              const char * buffer,
              size_t       count )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Check that we've got a reasonable buffer */

        if ( buffer == NULL ) {
                gpib_error   = GPIB_EARG;
                gpib_status |= GPIB_ERR;
                return gpib_status;
        }

        /* Nothing to be done if there are no characters to be send */

        if ( count == 0 ) {
                gpib_status |= GPIB_CMPL;
                gpib_error   = 0;
                gpib_count   = 0;
                return gpib_status;
        }

        /* Set eos character and mode and the timeout value for the device */

        if ( gpib_setup_eos( dev )  & GPIB_ERR ||
             gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        /* Send data to the device */

        return gpib_ioctl( GPIB_IOC_WRT, ADDRESS( dev->pad, dev->sad ),
                           buffer, count );
}


/*-------------------------------------------------------------*
 * Reads 'count' bytes into 'buffer' from the device addressed
 * by 'dev_dscr' (which was returned by a previous call of
 * gpib_find()). Before the data are read from the device the
 * EOS character and EOS mode for the device are set as well
 * as the timeout value.
 *-------------------------------------------------------------*/

int gpib_rd( int    dev_dscr,
             char * buffer,
             size_t count )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Check that we've got a reasonable buffer */

        if ( buffer == NULL ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Nothing to be done if there are no characters to be read */

        if ( count == 0 ) {
                gpib_status |= GPIB_CMPL;
                gpib_error   = 0;
                gpib_count   = 0;
                return gpib_status;
        }

        /* Set eos character and mode and the timeout value for the device */

        if ( gpib_setup_eos( dev ) & GPIB_ERR ||
             gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        /* Read the data from the device */

        return gpib_ioctl( GPIB_IOC_READ, ADDRESS( dev->pad, dev->sad ),
                           buffer, count );
}


/*--------------------------------------------------------------*
 * Writes a command of length 'count' passed to the function in
 * 'buffer' onto the GPIB bus. The first parameter has to be
 * the device descriptor of the master device, i.e. the card
 * controlling the bus.
 *--------------------------------------------------------------*/

int gpib_cmd( const char * buffer,
              size_t       count )
{
        Device *dev;


        if ( ! board_is_online ) {
                gpib_status &= ~ ( GPIB_CMPL | GPIB_ERR |
                                   GPIB_END | GPIB_TIMO );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EDVR;
                return gpib_status;
        }

        dev = &devices[ master_dev ];

        if ( gpib_setup_timo( dev ) != 0 )
                return gpib_status;

        /* Check that we got a reasonable buffer */

        if ( buffer == NULL ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Nothing to be done if there are no charcters to be written */

        if ( count == 0 ) {
                gpib_status |= GPIB_CMPL;
                gpib_error   = 0;
                gpib_count   = 0;
                return gpib_status;
        }

        /* Send the bytes to the bus */

        return gpib_ioctl( GPIB_IOC_CMD, buffer, count );
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
