/*
  $Id: gpib_list.c 673 2006-01-19 12:13:02Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This rogramm parses the GPIB configuration file, which is per default
  /etc/gpib.conf or, if the environment variable GPIB_CONF is set, the
  file defined by this variable.

  It returns, if called with no command line options or the option "-l",
  a list of the devices defined in this file starting with the decive
  name, the primary GPIB address, and, for the master device only, the
  name of the device file associated with the master device.

  If called with the option "-m" it returns a list of command line options
  suitable to be used with the mknod program to create device files for
  each device in the GPIB configuration file. Each line of the output
  consists of the decive file name, the character 'c', the major device
  number and the minor device number (identical to the primary GPIB address).
  The directory for the device files is the same as the one for the device
  file of the master device (which must be defined in the configuration
  file). The major device number is determined from the entry in 
  "/proc/devices" for the master device. For this to work the module for
  the GPIB device must already be loaded and the name of the device
  specified either as compile time option or via an option to insmod must
  be identical to the name given in the GPIB configuration file for the
  master device.

  If called with the option "-s" the programm simply prints out a short
  message indicating if the module is loaded or not (determined by checking
  "/proc/devices" for the name of the master device as given in the
  GPIB configuration file). If the module is loaded the return status of
  the program indicates success, otherwise failure.

  The GPIB configuration file may contain only one master device (controller).
  In both the "-l" as well as the "-m" mode of the program the entry for
  the master decice will always be listed as the first device. Devices
  that have a secondary GPIB address assigned are not listed at all. If
  there are devices having the same primary GPB address only the device
  appearing first in the GPIB configuration file will be listed (if there
  is a device having the same primary address as the master device it will
  never be listed but only the master device even if the entry for the
  master device appears only after the entry for the device with the same
  GPIB address).

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  To contact the author send email to: jt@toerring.de
*/


#include "lib.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


static Device devices[ GPIB_MAX_DEVICES ];
static int num_devices = 0;
static const char *cfg;


static int gpib_parse_config( const char *file );
static void do_checks( void );
int gpib_dev_setup( Device *temp_dev );
static void usage( void );

extern int yyparse( void );
extern int dev_line;


#define BUF_LEN 128


/*---------------------------------------------------------------------*
 *---------------------------------------------------------------------*/

int main( int    argc,
		  char * argv[ ] )
{
	int i;
	int flags = 0;
	char buffer[ BUF_LEN ];
	FILE *dlf;
	char name[ BUF_LEN ];
	int major;


	if ( argc > 2 )
		usage( );

	if ( argc > 1 )
	{
		switch ( argv[ 1 ][ 1 ] )
		{
			case 'l' :
				flags = 0;
				break;

			case 'm' :
				flags = 1;
				break;

			case 's' :
				flags = 2;
				break;

			case 'n' :
				flags = 3;
				break;

			case 'd' :
				flags = 4;
				break;

			default :
				usage( );
		}

		if ( argv[ 1 ][ 2 ] != '\0' )
			usage( );
	}

	for ( i = 0; i < GPIB_MAX_DEVICES; i++ )
	{
		devices[ i ].name[ 0 ] = '\0';
		devices[ i ].flags = 0;
		devices[ i ].file[ 0 ] = '\0';
		devices[ i ].ename[ 0 ] = '\0';
	}

	cfg = getenv( "GPIB_CONF" );
	if ( cfg == NULL )
		cfg = GPIB_DEF_CONFIG_FILE;
	if ( gpib_parse_config( cfg  ) != 0 )
		return EXIT_FAILURE;

	do_checks( );

	if ( flags == 0 )
		for ( i = 0; i < num_devices; i++ )
		{
			fprintf( stdout, "\"%s\" %d",
					 devices[ i ].name, devices[ i ].pad );
			if ( i == 0 )
				fprintf( stdout, " \"%s\"", devices[ i ].file );
			fprintf( stdout, "\n" );
		}
	else if ( flags == 3 )
		fprintf( stdout, "%s\n", devices[ 0 ].name );
	else if ( flags == 4 )
	{
		* strrchr( devices[ 0 ].file, '/' ) = '\0';
		fprintf( stdout, "%s\n", devices[ 0 ].file );
	}
	else
	{
		if ( ( dlf = fopen( "/proc/devices", "r" ) ) == NULL )
		{
			fprintf( stderr, "Can't open file \"/proc/devices\".\n" );
			exit( EXIT_FAILURE );
		}

		do
		{
			fgets( buffer, BUF_LEN, dlf );

			if ( buffer[ 0 ] == '\n' )
				break;

			if ( ! strcmp( buffer, "Character devices:\n" ) )
				continue;

			sscanf( buffer, "%d %s", &major, name );
		} while ( strcmp( name, devices[ 0 ].name ) );

		fclose( dlf );

		if ( strcmp( name, devices[ 0 ].name ) )
		{
			fprintf( stderr, "Device \"%s\" not found in \"/proc/devices\".\n",
					 devices[ 0 ].name );
			exit( EXIT_FAILURE );
		}

		if ( flags == 2 )
		{
			fprintf( stdout, "Device \"%s\" is up and running.\n",
					 devices[ 0 ].name );
			exit( EXIT_SUCCESS );
		}

		for ( i = 0; i < num_devices; i++ )
			fprintf( stdout, "\"%s\" c %d %d\n", devices[ i ].file, major,
					 devices[ i ].pad );
	}

	return EXIT_SUCCESS;
}


/*--------------------------------------------------------*
 * Function opens the configuration file, runs the parser
 * and finally checks the plausibility of the results.
 *--------------------------------------------------------*/

static int gpib_parse_config( const char * file )
{
	extern FILE *yyin;
	FILE *fp;
	int retval;


	if ( ( fp = fopen( file, "r" ) ) == NULL ) {
		fprintf( stderr, "Failed to open configuration file \"%s\".\n",
				 cfg );
		return -1;
	}

	yyin = fp;
	retval = yyparse( );
	fclose( fp );
	return retval;
}


/*---------------------------------------------------------------------*
 *---------------------------------------------------------------------*/

static void do_checks( void )
{
	int i, j;
	Device tmp_dev;
	size_t len;


	/* Make the master device the first device in the list, abort if there's
	   more than one controller */

	for ( i = 1; i < num_devices; i++ )
		if ( devices[ i ].flags & IS_MASTER )
		{
			if ( devices[ 0 ].flags & IS_MASTER )
			{
				fprintf( stderr, "More than one controller found in "
						 "configuration file \"%s\", first one is %s, the "
						 "other %s.\n",
						 cfg, devices[ 0 ].name, devices[ i ].name );
				exit( EXIT_FAILURE );
			}

			memcpy( &tmp_dev, devices, sizeof tmp_dev );
			memcpy( devices, devices + i, sizeof *devices );
			memcpy( devices + i, &tmp_dev, sizeof tmp_dev );
		}

	/* Throw out devices that have a secondary address set */

	for ( i = 0; i < num_devices; i++ )
	{
		if ( devices[ i ].sad == 0 )
			continue;

		memmove( devices + i, devices + i + 1,
				 ( num_devices - i - 1 ) * sizeof( *devices ) );
		num_devices--;
		i--;
	}

	/* Throw out devices that have the same primary address as one of the
	   devices listed earlier */

	for ( i = 0; i < num_devices; i++ )
		for ( j = 0; j < i; j++ )
		{
			if ( devices[ i ].pad != devices[ j ].pad )
				continue;

			if ( i == num_devices - 1 )
			{
				num_devices--;
				return;
			}

			memmove( devices + i, devices + i + 1,
					 ( num_devices - i - 1 ) * sizeof( *devices ) );
			num_devices--;
			j = -1;
		}

	/* Finally make up the file names for the non-controller devices */

	if ( devices[ 0 ].file[ 0 ] == '\0' )
	{
		fprintf( stderr, "Device file name for controller %s is missing in "
				 "configuration file \"%s\".\n", devices[ 0 ].name, cfg );
		exit ( EXIT_FAILURE );
	}

	if ( devices[ 0 ].file[ 0 ] != '/' )
	{
		fprintf( stderr, "Device file name for controller %s does not contain "
				 "an absolute path in configuration file \"%s\".\n",
				 devices[ 0 ].name, cfg );
		exit ( EXIT_FAILURE );
	}

	len = strrchr( devices[ 0 ].file, '/' ) - devices[ 0 ].file + 1;

	for ( i = 1; i < num_devices; i++ )
	{
		strncpy( devices[ i ].file, devices[ 0 ].file, len );
		strcpy( devices[ i ].file + len, devices[ i ].name );
	}
}


/*---------------------------------------------------------------------*
 * Function is called from the parser to check a few device parameters
 * and to copy them to the appropriate device structure.
 *---------------------------------------------------------------------*/

int gpib_dev_setup( Device * temp_dev )
{
	int i;


	/* Take care that there aren't too many devices */

	if ( num_devices >= GPIB_MAX_DEVICES )
		return -1;

	/* Check validity of name, addresses and timeout flag */

	if ( *temp_dev->name == '\0' ) {
		fprintf( stderr, "Name of device missing near line %d in "
			 "\"%s\".\n", dev_line, cfg );
		return -1;
	}

	if ( temp_dev->pad < 0 || temp_dev->pad > 0x1e ) {
		fprintf( stderr, "Invalid primary GPIB address %d for device "
			 "\"%s\" in \"%s\".\n", temp_dev->pad, temp_dev->name, cfg );
		return -1;
	}

	if ( temp_dev->sad != 0 &&
		 ( temp_dev->sad < 0x60 || temp_dev->sad > 0x7e ) ) {
		fprintf( stderr, "Invalid secondary GPIB address %d for device \"%s\" "
				 "in \"%s\".\n", temp_dev->sad, temp_dev->name, cfg );
		return -1;
	}

	/* Check that the device name is unique */

	for ( i = 0; i < num_devices; i++ )
		if ( ! strcmp( devices[ i ].name, temp_dev->name ) )
		{
			fprintf( stderr, "Device \"%s\" found more than once "
				 "in configuration file \"%s\".\n", temp_dev->name, cfg );
			return -1;
		}

	/* Copy to real device structure and increment device count */

	memcpy( devices + num_devices++, temp_dev, sizeof( Device ) );

	return 0;
}


/*---------------------------------------------------------------------*
 *---------------------------------------------------------------------*/

static void usage( void )
{
	fprintf( stderr,
			 "Usage: gpib_list [OPTION]\n"
			 "Returns list of the devices in the GPIB configuration file\n"
			 "(\"/etc/gpib.conf\" or set by environment variable GPIB_CONF).\n"
			 "  -l   list all devices with minor device number (default)\n"
			 "  -m   list device files etc. appropriate for use with 'mknod'\n"
			 "  -s   test if device module is loaded\n"
			 "  -h   display this text and exit\n"
		     "\n"
		     "Report bugs to <jt@toerring.de>.\n" );
	exit( EXIT_FAILURE );
}
